#!/usr/bin/env python

"""Defines an antenna's or entire radar system's radiation pattern,
also defines physical antennas for RX and TX.
"""

import numpy as np
from scipy.constants import speed_of_light
import spacecoords
import pyant
from sorts.types import Float_as_deg
from sorts import frames

StationId = int
"""A unique int16 that identifies a radar station"""

"""
A structured unique identifer for a station of a radar.

Intended to be used as the uid of a `Station` object.
"""


class Station(object):
    """A radar station.

    :param float lat: Geographical latitude of radar station in decimal degrees  (North+).
    :param float lon: Geographical longitude of radar station in decimal degrees (East+).
    :param float alt: Geographical altitude above geoid surface of radar station in meter.
    :param float min_elevation: Elevation threshold for the radar station in degrees,
        i.e. it cannot detect or point below this elevation.
    :param pyant.Beam beam: Radiation pattern for radar station.


    :ivar float lat: Geographical latitude of radar station in decimal degrees  (North+).
    :ivar float lon: Geographical longitude of radar station in decimal degrees (East+).
    :ivar float alt: Geographical altitude above geoid surface of radar station in meter.
    :ivar float min_elevation: Elevation threshold for the radar station in degrees,
        i.e. it cannot detect or point below this elevation.
    :ivar numpy.array ecef: The ECEF coordinates of the radar station
        calculated using :func:`frames.geodetic_to_ITRS`.
    :ivar pyant.Beam beam: Radiation pattern for radar station.
    :ivar bool enabled: Indicates if this station is turned on or off.

    """

    def __init__(
        self,
        lat,
        lon,
        alt,
        min_elevation: Float_as_deg,
        beam: pyant.beam.Beam,
        frequency: float | None = None,
        beam_parameters: pyant.types.Parameters | None = None,
        uid: StationId = 0,
    ):
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self.min_elevation = min_elevation
        self.ecef = frames.geodetic_to_ITRS(lat, lon, alt, degrees=True)
        ecef_lla = spacecoords.spherical.cart_to_sph(self.ecef, degrees=True)
        self.ecef_lat = ecef_lla[1]
        self.ecef_lon = 90 - ecef_lla[0]
        self.ecef_alt = ecef_lla[2]
        self.beam = beam
        self.enabled = True
        self.pointing_range = None
        self.uid = uid

        # TODO: attributes `frequency`, `wavelength` and `beam_parameters` are a tmp hack
        self.frequency = frequency
        self.wavelength = None if frequency is None else speed_of_light / frequency
        self.beam_parameters = beam_parameters

    def __repr__(self):
        return f"Station(uid={self.uid}, lat={self.lat}, lon={self.lon}, alt={self.alt})"

    def field_of_view(self, states, **kwargs):
        """Determines the field of view of the station.
        Should be vectorized over second dimension of states.
        Needs to return a numpy boolean array with `True` when the state is inside the FOV.

        Used to determine when a "pass" is occurring based on the input ECEF states and times.
        The default implementation is a minimum elevation check.
        """
        zenith = np.array([0, 0, 1], dtype=np.float64)

        enu = self.enu(states[:3, :])

        zenith_ang = spacecoords.linalg.vector_angle(zenith, enu, degrees=True)
        check = zenith_ang < 90.0 - self.min_elevation

        return check

    def rebase(self, lat, lon, alt):
        """Change geographical location of the station."""
        self.lat = lat
        self.lon = lon
        self.alt = alt
        self.ecef = frames.geodetic_to_ITRS(lat, lon, alt, degrees=True)
        ecef_lla = spacecoords.spherical.cart_to_sph(self.ecef, degrees=True)
        self.ecef_lat = ecef_lla[1]
        self.ecef_lon = 90 - ecef_lla[0]
        self.ecef_alt = ecef_lla[2]

    def copy(self):
        st = Station(
            lat=self.lat,
            lon=self.lon,
            alt=self.alt,
            min_elevation=self.min_elevation,
            beam=self.beam.copy(),
            frequency=self.frequency,
            beam_parameters=self.beam_parameters,
            uid=self.uid,
        )
        st.enabled = self.enabled
        return st

    def enu(self, ecefs):
        """Converts a set of ECEF states to local ENU coordinates using geocentric zenith."""
        rel_ = ecefs.copy()
        rel_[:3, :] = rel_[:3, :] - self.ecef[:, None]
        rel_[:3, :] = frames.ecef_to_enu(
            self.ecef_lat,
            self.ecef_lon,
            self.ecef_alt,
            rel_[:3, :],
            degrees=True,
        )
        if ecefs.shape[0] > 3:
            rel_[3:, :] = frames.ecef_to_enu(
                self.ecef_lat,
                self.ecef_lon,
                self.ecef_alt,
                rel_[3:, :],
                degrees=True,
            )
        return rel_

    def point(self, k):
        """Point Station beam in local ENU coordinates using geocentric zenith."""
        self.beam.point(k)

    def point_ecef(self, point):
        """Point Station beam in location of ECEF coordinate. Returns local pointing direction."""
        k = frames.ecef_to_enu(
            self.ecef_lat,
            self.ecef_lon,
            self.ecef_alt,
            point,
            degrees=True,
        )
        k_norm = np.linalg.norm(k, axis=0)

        self.pointing_range = k_norm

        k = k / k_norm
        self.beam.point(k)
        return k

    @property
    def pointing(self):
        """Station beam pointing in local ENU coordinates using geocentric zenith."""
        return self.beam.pointing.copy()

    @property
    def pointing_ecef(self):
        """Station beam pointing in local ENU coordinates using geocentric zenith."""
        return frames.enu_to_ecef(
            self.ecef_lat,
            self.ecef_lon,
            self.ecef_alt,
            self.beam.pointing,
            degrees=True,
        )


class RX(Station):
    """A radar receiving system.

    :param float noise: Receiver noise in Kelvin, i.e. system temperature.

    :ivar float noise: Receiver noise in Kelvin, i.e. system temperature.
    """

    def __init__(
        self,
        lat,
        lon,
        alt,
        min_elevation,
        beam,
        noise,
        frequency: float | None = None,
        beam_parameters: pyant.types.Parameters | None = None,
        uid: StationId = 1,
    ):
        super().__init__(
            lat=lat,
            lon=lon,
            alt=alt,
            min_elevation=min_elevation,
            beam=beam,
            frequency=frequency,
            beam_parameters=beam_parameters,
            uid=uid,
        )
        self.noise = noise

    def copy(self):
        st = RX(
            lat=self.lat,
            lon=self.lon,
            alt=self.alt,
            min_elevation=self.min_elevation,
            beam=self.beam.copy(),
            noise=self.noise,
            frequency=self.frequency,
            beam_parameters=self.beam_parameters,
            uid=self.uid,
        )
        st.enabled = self.enabled
        return st


class TX(Station):
    """A transmitting radar station

    :param float bandwidth: Transmissions bandwidth.
    :param float duty_cycle: Maximum duty cycle, i.e. fraction of time
        transmission can occur at maximum power.
    :param float power: Transmissions power in watts.
    :param float pulse_length: Length of transmission pulse.
    :param float ipp: Time between consecutive pulses.
    :param int n_ipp: Number of pulses to coherently integrate.

    :ivar float bandwidth: Transmissions bandwidth.
    :ivar float duty_cycle: Maximum duty cycle, i.e. fraction of time
        transmission can occur at maximum power.
    :ivar float power: Transmissions power in watts.
    :ivar float pulse_length: Length of transmission pulse.
    :ivar float ipp: Time between consecutive pulses.
    :ivar int n_ipp: Number of pulses to coherently integrate.
    :ivar float coh_int_bandwidth: Effective bandwidth of receiver noise after coherent integration.
    """

    def __init__(
        self,
        lat,
        lon,
        alt,
        min_elevation,
        beam,
        power,
        bandwidth,
        duty_cycle,
        pulse_length=1e-3,
        ipp=10e-3,
        n_ipp=20,
        frequency: float | None = None,
        beam_parameters: pyant.types.Parameters | None = None,
        uid: StationId = 0,
    ):
        super().__init__(
            lat=lat,
            lon=lon,
            alt=alt,
            min_elevation=min_elevation,
            beam=beam,
            frequency=frequency,
            beam_parameters=beam_parameters,
            uid=uid,
        )

        self.bandwidth = bandwidth
        self.duty_cycle = duty_cycle
        self.power = power
        self.pulse_length = pulse_length
        self.ipp = ipp
        self.n_ipp = n_ipp
        self.coh_int_bandwidth = 1.0 / (pulse_length * n_ipp)

    def copy(self):
        st = TX(
            lat=self.lat,
            lon=self.lon,
            alt=self.alt,
            min_elevation=self.min_elevation,
            beam=self.beam.copy(),
            power=self.power,
            bandwidth=self.bandwidth,
            duty_cycle=self.duty_cycle,
            pulse_length=self.pulse_length,
            ipp=self.ipp,
            n_ipp=self.n_ipp,
            frequency=self.frequency,
            beam_parameters=self.beam_parameters,
            uid=self.uid,
        )
        st.enabled = self.enabled
        return st



#!/usr/bin/env python

"""

"""

import logging
import numpy as np
import scipy.constants

from .scheduler import Scheduler

from .. import signals
from ..passes import Pass

logger = logging.getLogger(__name__)

class ObservedParameters(Scheduler):
    """Bi-static radar observation parameters of hard targets.

    **Parameters calculated**

        * time
        * signal to noise ratio
        * range
        * range rate
        * transmitter local pointing to target k
        * receiver pointing to target k
        * radar cross section

        #TODO: Docstring

    """

    def __init__(self, radar, **kwargs):
        super().__init__(
            radar=radar,
        )

    def calculate_observation_jacobian(
        self, txrx_pass, space_object, variables, deltas, transforms={}, **kwargs
    ):
        """Calculate the observation and its Jacobean of a pass of a specific space object given the current state of the Scheduler.

        The Jacobean assumes that the SpaceObject has a Orbit state. To perturb non Orbit states a custom implementation is needed.

        NOTE: During the numerical calculation of the Jacobean only the range and range rates are calculated and `calculate_snr=False`.

        #TODO: Docstring
        """

        logger.debug(
            f"Obs.Param.:calculate_observation_jacobian:{variables}, deltas={deltas}"
        )

        t, generator = self(txrx_pass.start(), txrx_pass.end())
        if generator is None:
            return None, None

        data0 = self.calculate_observation(txrx_pass, t, generator, space_object, **kwargs)

        if data0 is None:
            return None, None

        keep = np.full(t.shape, False, dtype=bool)
        keep[data0["kept"]] = True

        r = np.empty((len(t), 2), dtype=np.float64)
        r_dot = np.empty((len(t), 2), dtype=np.float64)

        r[keep, 0] = data0["range"]
        r_dot[keep, 0] = data0["range_rate"]

        J = np.zeros([len(t) * 2, len(variables)], dtype=np.float64)

        kwargs["snr_limit"] = False
        kwargs["calculate_snr"] = False
        for ind, var in enumerate(variables):
            dso = space_object.copy()
            if var in transforms:
                Tx = transforms[var][0](getattr(dso, var)) + deltas[ind]
                dx = transforms[var][1](Tx)
            else:
                dx = getattr(dso, var) + deltas[ind]

            dso.update(**{var: dx})

            ddata = self.calculate_observation(txrx_pass, t, generator, dso, **kwargs)

            dkeep = np.full(t.shape, False, dtype=bool)
            dkeep[ddata["kept"]] = True
            keep = np.logical_and(keep, dkeep)

            r[dkeep, 1] = ddata["range"]
            r_dot[dkeep, 1] = ddata["range_rate"]

            dr = (r[:, 1] - r[:, 0]) / deltas[ind]
            dv = (r_dot[:, 1] - r_dot[:, 0]) / deltas[ind]

            J[: len(t), ind] = dr
            J[len(t) :, ind] = dv

        for key in data0:
            if key in ["kept"]:
                continue
            elif isinstance(data0[key], np.ndarray):
                data0[key] = data0[key][..., keep[data0["kept"]]]
            else:
                data0[key] = [x for ind_, x in enumerate(data0[key]) if keep[data0["kept"]][ind_]]
        data0["kept"] = np.argwhere(keep).flatten()

        Jkeep = np.full((len(t) * 2,), False, dtype=bool)
        Jkeep[: len(t)] = keep
        Jkeep[len(t) :] = keep

        J = J[Jkeep, :]

        return data0, J

    def get_beam_gain_and_wavelength(self, beam, enu, meta):
        """Given the input beam configured by the controller and the local (for that beam) coordinates of the observed object, get the correct gain and wavelength used for SNR calculation.

        The default is a maximum calculation based on pointing, this is used for e.g. RX digital beam-forming.
        """
        if len(beam.pointing.shape) > 1:
            g = np.max(
                [beam.gain(enu, ind={"pointing": pi}) for pi in range(beam.pointing.shape[1])]
            )
        else:
            g = beam.gain(enu)

        # TODO: ugly hack because we are gonna scrap this method anyway
        # this should to picked based on experiment setup
        lam = beam.wavelength[0]
        return g, lam

    def stop_condition(self, t, radar, meta, tx_enu, rx_enu, tx_index, rx_index):
        """Implement this function if there should be abort condition in the sequential generator. Work with both vectorization and linear iteration (as vectorization first caches the generator)."""
        return False

    def observable_filter(self, t, radar, meta, tx_enu, rx_enu, tx_index, rx_index):
        """Determines if the object was observable or not by the radar system"""
        return True

    def vectorized_get_beam_gain_and_wavelength(
        self, t, vectorized_data, tx_enus, rx_enus, tx_index, rx_index
    ):
        tx_g = self.radar.tx[tx_index].beam.gain(
            tx_enus,
            pointing=vectorized_data[:, 0:3].T,
            vectorized_parameters=True,
        )
        rx_g = self.radar.rx[rx_index].beam.gain(
            rx_enus,
            pointing=vectorized_data[:, 4:7].T,
            vectorized_parameters=True,
        )

        return tx_g, vectorized_data[:, 3], rx_g, vectorized_data[:, 7]

    def vectorized_observable_filter(
        self, t, vectorized_data, tx_enus, rx_enus, tx_index, rx_index
    ):
        return np.full(t.shape, True, dtype=bool)

    def get_vectorized_row(self, radar, meta, tx_index, rx_index):
        """Used to extract the used data from the `radar` instance when vectorizing `calculate_observation`.
        Input to `vectorized_observable_filter` and `vectorized_get_beam_gain_and_wavelength`.

        Should return a numpy vector, will be stored as a row-vector in the matrix passed to the vectorized functions.
        """
        row = np.empty((8,), dtype=np.float64)
        row[0:3] = radar.tx[tx_index].beam.pointing[:]
        row[3] = radar.tx[tx_index].beam.wavelength
        row[4:7] = radar.rx[rx_index].beam.pointing[:]
        row[7] = radar.rx[rx_index].beam.wavelength
        return row

    def extend_meta(self, t, txi, rxi, radar, meta):
        meta["pulse_length"] = radar.tx[txi].pulse_length
        meta["ipp"] = radar.tx[txi].ipp
        meta["n_ipp"] = radar.tx[txi].n_ipp

    def calculate_observation(
        self,
        txrx_pass,
        t,
        generator,
        space_object,
        epoch=None,
        calculate_snr=True,
        doppler_spread_integrated_snr=False,
        interpolator=None,
        snr_limit=True,
        save_states=False,
        vectorize=False,
        extended_meta=True,
        blind_ranges=True,
    ):
        """Calculate the observation of a pass of a specific space object given the current state of the Scheduler.

        #ASSUMES INPUT t IS RELATIVE SPACE OBJECT EPOCH unless epoch is given

        #TODO: Docstring
        """

        txi, rxi = txrx_pass.station_id

        logger.debug(
            f"Obs.Param.:calculate_observation:(tx={txi}, rx={rxi}), len(t) = {len(t)}"
        )

        if calculate_snr:
            diam = space_object.d
        else:
            diam = None

        spin_period = space_object.properties.get("spin_period", None)
        radar_albedo = space_object.properties.get("radar_albedo", 1.0)

        # t is always in scheduler relative time
        # t_samp is in space object relative time if there is a scheduler epoch, otherwise it is assumed that the epoch are the same
        # if there is an interpolator it is assumed that interpolation is done in space object relative time
        if epoch is None:
            t_samp = t
        else:
            t_samp = t + (epoch - space_object.epoch).sec

        if interpolator is not None:
            states = interpolator.get_state(t_samp)
        else:
            states = space_object.get_state(t_samp)

        snr = np.empty((len(t),), dtype=np.float64)
        snr_inch = np.empty((len(t),), dtype=np.float64)
        rcs = np.empty((len(t),), dtype=np.float64)
        keep = np.full((len(t),), True, dtype=bool)

        enus = [
            self.radar.tx[txi].enu(states),
            self.radar.rx[rxi].enu(states),
        ]
        ranges = [Pass.calculate_range(enu) for enu in enus]
        range_rates = [Pass.calculate_range_rate(enu) for enu in enus]

        metas = []

        if vectorize:
            powers = np.empty((len(t),), dtype=np.float64)
            t_slices = np.empty((len(t),), dtype=np.float64)
            pulse_lengths = np.empty((len(t),), dtype=np.float64)
            ipps = np.empty((len(t),), dtype=np.float64)
            bandwidths = np.empty((len(t),), dtype=np.float64)
            duty_cycles = np.empty((len(t),), dtype=np.float64)
            rx_noise_temps = np.empty((len(t),), dtype=np.float64)
            txrx_on = np.full((len(t),), False, dtype=bool)

            vectorized_data = None
            for ri, (radar, meta) in enumerate(generator):
                if extended_meta:
                    self.extend_meta(t[ri], txi, rxi, radar, meta)

                stop = self.stop_condition(
                    t[ri],
                    radar,
                    meta,
                    enus[0][:3, ri],
                    enus[1][:3, ri],
                    txi,
                    rxi,
                )
                if stop:
                    break

                metas.append(meta)
                vec_row = self.get_vectorized_row(radar, meta, txi, rxi)
                if vectorized_data is None:
                    vectorized_data = np.empty((len(t), len(vec_row)), dtype=vec_row.dtype)

                vectorized_data[ri, :] = vec_row

                t_slice_ = meta.get("t_slice", None)
                if t_slice_ is not None:
                    t_slices[ri] = t_slice_
                else:
                    t_slices[ri] = np.nan
                pulse_lengths[ri] = radar.tx[txi].pulse_length
                ipps[ri] = radar.tx[txi].ipp
                powers[ri] = radar.tx[txi].power
                bandwidths[ri] = radar.tx[txi].coh_int_bandwidth
                duty_cycles[ri] = radar.tx[txi].duty_cycle
                rx_noise_temps[ri] = radar.rx[rxi].noise

                if radar.tx[txi].enabled and radar.rx[rxi].enabled:
                    txrx_on[ri] = True

            observable = self.vectorized_observable_filter(
                t,
                vectorized_data,
                enus[0][:3, :],
                enus[1][:3, :],
                txi,
                rxi,
            )
            keep = np.logical_and(observable, txrx_on)

            if calculate_snr:
                (
                    tx_g,
                    tx_wavelength,
                    rx_g,
                    rx_wavelength,
                ) = self.vectorized_get_beam_gain_and_wavelength(
                    t[keep],
                    vectorized_data[keep, :],
                    enus[0][:3, keep],
                    enus[1][:3, keep],
                    txi,
                    rxi,
                )

                if doppler_spread_integrated_snr:
                    snr[keep], snr_inch[keep] = signals.doppler_spread_hard_target_snr(
                        t_slices[keep],
                        spin_period,
                        tx_g,
                        rx_g,
                        tx_wavelength,
                        powers[keep],
                        ranges[0][keep],
                        ranges[1][keep],
                        duty_cycle=duty_cycles[keep],
                        diameter=diam,
                        bandwidth=bandwidths[keep],
                        rx_noise_temp=rx_noise_temps[keep],
                        radar_albedo=radar_albedo,
                    )
                else:
                    snr[keep] = signals.hard_target_snr(
                        tx_g,
                        rx_g,
                        tx_wavelength,
                        powers[keep],
                        ranges[0][keep],
                        ranges[1][keep],
                        diameter=diam,
                        bandwidth=bandwidths[keep],
                        rx_noise_temp=rx_noise_temps[keep],
                        radar_albedo=radar_albedo,
                    )

                if blind_ranges:
                    snr_modulation = np.ones((len(t),), dtype=np.float64)
                    for ch_txi, ch_rxi in self.radar.joint_stations:
                        if ch_rxi == rxi:
                            delay = (ranges[0][keep] + ranges[1][keep]) / scipy.constants.c
                            ipp_f = np.mod(delay, ipps[keep])

                            inds = ipp_f <= pulse_lengths[keep]
                            snr_modulation[keep][inds] = pulse_lengths[keep][inds]

                            inds = ipp_f >= ipps[keep] - pulse_lengths[keep]
                            snr_modulation[keep][inds] = (
                                ipps[keep][inds] - ipp_f[inds]
                            ) / pulse_lengths[keep][inds]

                            break
                    snr[keep] = snr[keep] * snr_modulation[keep]

                rcs[keep] = signals.hard_target_rcs(
                    wavelength=tx_wavelength,
                    diameter=diam,
                )
                if snr_limit:
                    snr_db = np.empty(snr.shape, dtype=np.float64)
                    snr_db[snr < 1e-9] = np.nan
                    snr_db[snr >= 1e-9] = np.log10(snr[snr >= 1e-9]) * 10.0

                    keep[np.logical_or(np.isnan(snr_db), np.isinf(snr_db))] = False
                    keep[keep] = snr_db[keep] > radar.min_SNRdb

        else:
            for ti, (radar, meta) in enumerate(generator):
                if extended_meta:
                    self.extend_meta(t[ti], txi, rxi, radar, meta)

                stop = self.stop_condition(
                    t[ti],
                    radar,
                    meta,
                    enus[0][:3, ti],
                    enus[1][:3, ti],
                    txi,
                    rxi,
                )
                if stop:
                    keep[ti:] = False
                    break

                metas.append(meta)

                observable = self.observable_filter(
                    t[ti],
                    radar,
                    meta,
                    enus[0][:3, ti],
                    enus[1][:3, ti],
                    txi,
                    rxi,
                )

                if not (radar.tx[txi].enabled and radar.rx[rxi].enabled and observable):
                    keep[ti] = False
                    continue

                if calculate_snr:
                    snr_modulation = 1.0
                    if blind_ranges:
                        # check if target is in radars blind range
                        # assume synchronized transmitting
                        # assume decoding of partial pulses is possible and linearly decreases signal strength
                        for ch_txi, ch_rxi in radar.joint_stations:
                            if ch_rxi == rxi:
                                delay = (ranges[0][ti] + ranges[1][ti]) / scipy.constants.c
                                ipp_f = np.mod(delay, radar.tx[txi].ipp)

                                if ipp_f <= radar.tx[txi].pulse_length:
                                    snr_modulation = ipp_f / radar.tx[txi].pulse_length
                                elif ipp_f >= radar.tx[txi].ipp - radar.tx[txi].pulse_length:
                                    snr_modulation = (radar.tx[txi].ipp - ipp_f) / radar.tx[
                                        txi
                                    ].pulse_length
                                break

                    tx_g, tx_wavelength = self.get_beam_gain_and_wavelength(
                        radar.tx[txi].beam,
                        enus[0][:3, ti],
                        meta,
                    )
                    rx_g, rx_wavelength = self.get_beam_gain_and_wavelength(
                        radar.rx[rxi].beam,
                        enus[1][:3, ti],
                        meta,
                    )

                    if doppler_spread_integrated_snr:
                        snr[ti], snr_inch[ti] = signals.doppler_spread_hard_target_snr(
                            meta.get("t_slice", np.nan),
                            spin_period,
                            tx_g,
                            rx_g,
                            tx_wavelength,
                            radar.tx[txi].power,
                            ranges[0][ti],
                            ranges[1][ti],
                            duty_cycle=radar.tx[txi].duty_cycle,
                            diameter=diam,
                            bandwidth=radar.tx[txi].coh_int_bandwidth,
                            rx_noise_temp=radar.rx[rxi].noise,
                            radar_albedo=radar_albedo,
                        )
                    else:
                        snr[ti] = signals.hard_target_snr(
                            tx_g,
                            rx_g,
                            tx_wavelength,
                            radar.tx[txi].power,
                            ranges[0][ti],
                            ranges[1][ti],
                            diameter=diam,
                            bandwidth=radar.tx[txi].coh_int_bandwidth,
                            rx_noise_temp=radar.rx[rxi].noise,
                        )
                    snr[ti] *= snr_modulation

                    rcs[ti] = signals.hard_target_rcs(
                        wavelength=tx_wavelength,
                        diameter=diam,
                    )
                    if snr_limit:
                        if snr[ti] < 1e-9:
                            snr_db = np.nan
                        else:
                            snr_db = np.log10(snr[ti]) * 10.0

                        if np.isnan(snr_db) or np.isinf(snr_db):
                            keep[ti] = False
                        else:
                            keep[ti] = snr_db > radar.min_SNRdb

                else:
                    snr[ti] = np.nan
                    rcs[ti] = np.nan

        data = dict(
            t=t,
            snr=snr,
            range=ranges[0] + ranges[1],
            range_rx=ranges[1],
            range_rate=range_rates[0] + range_rates[1],
            tx_k=enus[0][:3, :] / ranges[0],
            rx_k=enus[1][:3, :] / ranges[1],
            rcs=rcs,
            metas=metas,
        )
        if doppler_spread_integrated_snr:
            data["snr_inch"] = snr_inch
        if save_states:
            data["states"] = states

        if np.any(keep):
            for key in data:
                if isinstance(data[key], np.ndarray):
                    data[key] = data[key][..., keep]
                else:
                    data[key] = [x for ind_, x in enumerate(data[key]) if keep[ind_]]
            data["kept"] = np.argwhere(keep).flatten()
        else:
            data = None

        logger.debug(f"Obs.Param.:calculate_observation:complete")
        return data



#!/usr/bin/env python

"""wrapper for the SGP4 propagator"""
from typing import Any
import logging
from dataclasses import dataclass
import numpy as np
from astropy.time import Time, TimeDelta
import scipy.optimize
import pyorb
import sgp4
from sgp4.api import Satrec, SGP4_ERRORS
import sgp4.earth_gravity
import spacecoords.celestial as cel

from sorts.utils import convert_to_relative_time
from sorts.space_object import SpaceObject
from sorts.types import Settings, GravModels, NDArray_6, Frames, NDArray_6xN, NDArray_N
from .base import Propagator

logger = logging.getLogger(__name__)


def _sgp4_elems2cart(
    kep: NDArray_6xN | NDArray_6,
    grav_model,
) -> NDArray_6xN | NDArray_6:
    """Orbital elements to cartesian coordinates.
    Wrap pyorb-function to use mean anomaly, km and reversed order on aoe and raan.
    Output in SI.

    Neglecting mass is sufficient for this calculation
    (the standard gravitational parameter is 24 orders larger then the change).
    """
    _kep = kep.copy()
    _kep[0, ...] *= 1e3
    tmp = _kep[4, ...].copy()
    _kep[4, ...] = _kep[3, ...]
    _kep[3, ...] = tmp
    _kep[5, ...] = pyorb.mean_to_true(_kep[5, ...], _kep[1, ...], degrees=False)
    cart = pyorb.kep_to_cart(_kep, mu=grav_model.mu * 1e9, degrees=False)
    return cart


def _cart2sgp4_elems(
    cart: NDArray_6xN | NDArray_6,
    grav_model,
    degrees: bool = False,
) -> NDArray_6xN | NDArray_6:
    """Cartesian coordinates to orbital elements.
    Wrap pyorb-function to use mean anomaly, km and reversed order on aoe and raan.

    Neglecting mass is sufficient for this calculation
    (the standard gravitational parameter is 24 orders larger then the change).
    """
    kep = pyorb.cart_to_kep(cart, mu=grav_model.mu * 1e9, degrees=False)
    kep[0, ...] *= 1e-3
    tmp = kep[4, ...].copy()
    kep[4, ...] = kep[3, ...]
    kep[3, ...] = tmp
    kep[5, ...] = pyorb.true_to_mean(kep[5, ...], kep[1, ...], degrees=False)
    return kep


def kep_to_mean_elements(orb: pyorb.Orbit, degrees: bool = False) -> NDArray_6xN:
    """Orbital elements to the convention used by mean elements. Input assumes SI units.

    Neglecting mass is sufficient for this calculation
    (the standard gravitational parameter is 24 orders larger then the change).
    """
    kep = np.empty_like(orb._kep)
    kep[0, ...] = orb.a * 1e-3
    kep[1, ...] = orb.e
    kep[2, ...] = orb.i
    kep[3, ...] = orb.Omega
    kep[4, ...] = orb.omega
    kep[5, ...] = orb.mean_anomaly

    if orb.degrees and not degrees:
        kep[2:, :] = np.radians(kep[2:, :])
    elif not orb.degrees and degrees:
        kep[2:, :] = np.degrees(kep[2:, :])

    return kep


def get_TLE_parameters(
    line1: str, line2: str, gravity_model: GravModels = "WGS84"
) -> dict[str, Any]:
    line1, line2 = line_decode(line1), line_decode(line2)

    grav_ind = getattr(sgp4.api, gravity_model.upper())
    satellite = Satrec.twoline2rv(line1, line2, grav_ind)
    ret = {}
    for key in ["bstar", "satnum", "jdsatepochF", "jdsatepoch"]:
        ret[key] = getattr(satellite, key)
    return ret


def line_decode(line: str | np.bytes_) -> str:
    if isinstance(line, np.bytes_):
        rline = str(line.astype("U"))
    elif not isinstance(line, str):
        try:
            rline = line.decode()
        except (UnicodeDecodeError, AttributeError):
            pass
    else:
        rline = line

    return rline


def get_B(properties):
    # TODO: we need a way to specify which properties can exist and what they are somehow....
    if "B" in properties:
        B = properties["B"]
    elif "area_to_mass" in properties:
        B = 0.5 * properties.get("C_D", 2.3) * properties["area_to_mass"]
    elif "A" in properties and "m" in properties:
        B = 0.5 * properties.get("C_D", 2.3) * properties["A"] / properties["m"]
    elif "d" in properties and "m" in properties:
        A = np.pi * (properties["d"] / 2) ** 2
        B = 0.5 * properties.get("C_D", 2.3) * A / properties["m"]
    else:
        B = 0.0
    return B


@dataclass
class Sgp4Settings(Settings):
    out_frame: Frames = "TEME"
    gravity_model: GravModels = "WGS84"
    mean_elements_input: bool = False
    tol: float = 1e-5
    tol_v: float = 1e-7
    sample_space_object_kepler_orbit: bool = False
    kepler_samples: int = 100
    kepler_extent: float = 0.2
    teme_to_tle_max_iter: int = 300
    teme_to_tle_minimize_start_samples: int = 1
    teme_to_tle_minimize_start_stds: tuple[float, ...] = (10.0, 0.01, 1.0, 2.0, 2.0, 2.0)
    teme_to_tle_minimize_bounds: tuple[tuple[float, float], ...] = (
        (6371.0, np.inf),
        (0, 1),
        (0, np.pi),
        (0, 2 * np.pi),
        (0, 2 * np.pi),
        (0, 2 * np.pi),
    )


class Sgp4(Propagator[Sgp4Settings]):
    """Propagator class implementing the SGP4 propagator."""

    def __init__(self, settings: Sgp4Settings):
        super().__init__(settings=settings)

        self.sgp4_mjd0 = Time("1949-12-31 00:00:00", format="iso", scale="ut1").mjd
        self.rho0 = 2.461e-5 / 6378.135e3  # kg/m^2/m

    def propagate_tle(
        self,
        line1: str,
        line2: str,
        times: Time | TimeDelta | NDArray_N,
    ) -> NDArray_6xN:
        """Propagate a TLE pair"""

        grav_ind = getattr(sgp4.api, self.settings.gravity_model.upper())
        # grav_model = getattr(sgp4.earth_gravity, self.settings.gravity_model.lower())
        line1, line2 = line_decode(line1), line_decode(line2)

        satellite = Satrec.twoline2rv(line1, line2, grav_ind)

        epoch = Time(satellite.jdsatepoch + satellite.jdsatepochF, format="jd", scale="utc")
        logger.debug(f"SGP4:propagate_tle:epoch={epoch}")
        tv = convert_to_relative_time(epoch, times)
        td = TimeDelta(tv, format="sec")

        times = epoch + td

        jd_f = times.jd2
        jd0 = times.jd1

        if isinstance(jd_f, float) or isinstance(jd_f, int):
            states = np.empty((6,), dtype=np.float64)
            error, r, v = satellite.sgp4(jd0, jd_f)
            states[:3] = r
            states[3:] = v
            errors = [error]
        else:
            states = np.empty((6, jd_f.size), dtype=np.float64)
            errors, r, v = satellite.sgp4_array(jd0, jd_f)
            states[:3, ...] = r.T
            states[3:, ...] = v.T

        for ind, err in enumerate(errors):
            if err != 0:
                logger.error(f"SGP4:propagate_tle:step-{ind}:{SGP4_ERRORS[err]}")

        states *= 1e3  # km to m, km/s to m/s

        if self.settings.out_frame != "TEME":
            states = cel.convert(
                times,
                states,
                in_frame="TEME",
                out_frame=self.settings.out_frame,
                frame_kwargs={},
            )

        return states

    def propagate(
        self,
        space_object: SpaceObject,
        times: Time | TimeDelta | NDArray_N,
    ) -> NDArray_6xN:
        """Propagate a state

        `space_object` properties only information needed for ballistic coefficient
        `B` used by SGP4. Either `B` or `C_D`, `A` and `m` must be supplied.

        - B: Ballistic coefficient
        - C_D: Drag coefficient
        - A: Cross-sectional Area
        - m: Mass
        """
        logger.debug("SGP4:propagate")
        if space_object.state.num > 1:
            t_samps = space_object.properties["state_sample_times"]
        else:
            t_samps = None

        tv = convert_to_relative_time(space_object.epoch, times)
        td = TimeDelta(tv, format="sec")
        t = space_object.epoch + td

        sgp4_epoch = space_object.epoch.mjd - self.sgp4_mjd0

        B = get_B(space_object.properties)
        logger.debug(f"SGP4:propagate:B = {B}")
        state0 = space_object.state.copy()

        if space_object.frame != "TEME":
            state0._cart = cel.convert(
                space_object.epoch,
                state0._cart,
                in_frame=space_object.frame,
                out_frame="TEME",
                frame_kwargs={},
            )
            state0.calculate_kepler()

        if self.settings.mean_elements_input:
            mean_elements = kep_to_mean_elements(state0, degrees=False)
            assert mean_elements.size == 6, "Can not propagate multiple objects"
            if len(mean_elements.shape) > 2:
                mean_elements.shape = (mean_elements.size,)
        else:
            if self.settings.sample_space_object_kepler_orbit:
                mean_elements = self.space_object_to_mean_elements(
                    space_object,
                )
            else:
                mean_elements = self.TEME_to_TLE(
                    state0._cart,
                    t=t_samps,
                    epoch=space_object.epoch,
                    B=B,
                    tol=self.settings.tol,
                    tol_v=self.settings.tol_v,
                )

        if np.any(np.isnan(mean_elements)):
            raise Exception("Could not compute SGP4 initial state: {}".format(mean_elements))

        states = self.propagate_mean_elements(
            t.jd1,
            t.jd2,
            mean_elements,
            sgp4_epoch,
            B,
        )

        if self.settings.out_frame != "TEME":
            states = cel.convert(
                t,
                states,
                in_frame="TEME",
                out_frame=self.settings.out_frame,
                frame_kwargs={},
            )

        logger.debug("SGP4:propagate:completed")

        return states

    def get_mean_elements(
        self,
        line1: str,
        line2: str,
        radians: bool = False,
    ) -> tuple[NDArray_6, float, Time]:
        """Extract the mean elements in SI units (a [m], e [1], inc [deg],
        raan [deg], aop [deg], mu [deg]), B-parameter (not bstar) and epoch
        from a two line element pair.
        """

        grav_ind = getattr(sgp4.api, self.settings.gravity_model.upper())
        grav_model = getattr(sgp4.earth_gravity, self.settings.gravity_model.lower())
        line1, line2 = line_decode(line1), line_decode(line2)

        xpdotp = 1440.0 / (2.0 * np.pi)  # 229.1831180523293

        satrec = Satrec.twoline2rv(line1, line2, grav_ind)

        B = satrec.bstar / (grav_model.radiusearthkm * 1e3) * 2 / self.rho0

        epoch = Time(satrec.jdsatepoch + satrec.jdsatepochF, format="jd", scale="utc")

        mean_elements = np.zeros((6,), dtype=np.float64)

        n0 = satrec.no_kozai * xpdotp / (86400.0 / (2 * np.pi))

        mean_elements[0] = (np.sqrt(grav_model.mu) / n0) ** (2.0 / 3.0) * 1e3
        mean_elements[1] = satrec.ecco
        mean_elements[2] = satrec.inclo
        mean_elements[3] = satrec.nodeo
        mean_elements[4] = satrec.argpo
        mean_elements[5] = satrec.mo
        if not radians:
            mean_elements[2:] = np.degrees(mean_elements[2:])

        return mean_elements, B, epoch

    def propagate_mean_elements(
        self,
        jd0: NDArray_N | float,
        jd_f: NDArray_N | float,
        mean_elements: NDArray_6,
        sgp4_epoch: float,
        B: float,
    ) -> NDArray_6xN | NDArray_6:
        """Propagate sgp4 mean elements."""

        grav_ind = getattr(sgp4.api, self.settings.gravity_model.upper())
        grav_model = getattr(sgp4.earth_gravity, self.settings.gravity_model.lower())
        # Compute ballistic coefficient
        bstar = 0.5 * B * self.rho0  # B* in [1/m] using Density at q0[kg/m^3]
        n0 = np.sqrt(grav_model.mu) / ((mean_elements[0]) ** 1.5)

        # Scaling
        n0 = n0 * (86400.0 / (2 * np.pi))  # Convert to [rev/d]
        bstar = bstar * (grav_model.radiusearthkm * 1e3)  # Convert from [1/m] to [1/R_EARTH]

        satellite = Satrec()
        satellite.sgp4init(
            grav_ind,  # gravity model
            "i",  # 'a' = old AFSPC mode, 'i' = improved mode
            42,  # satnum: Satellite number
            sgp4_epoch,  # epoch: days since 1949 December 31 00:00 UT
            bstar,  # bstar: drag coefficient (/earth radii)
            0.0,  # [IGNORED BY SGP4] ndot: ballistic coefficient (revs/day)
            0.0,  # [IGNORED BY SGP4] nddot: second derivative of mean motion (revs/day^3)
            mean_elements[1],  # ecco: eccentricity
            mean_elements[4],  # argpo: argument of perigee (radians)
            mean_elements[2],  # inclo: inclination (radians)
            mean_elements[5],  # mo: mean anomaly (radians)
            n0 / (1440.0 / (2.0 * np.pi)),  # no_kozai: mean motion (radians/minute)
            mean_elements[3],  # nodeo: right ascension of ascending node (radians)
        )

        if isinstance(jd_f, float) or isinstance(jd_f, int):
            states = np.empty((6,), dtype=np.float64)
            error, r, v = satellite.sgp4(jd0, jd_f)
            states[:3] = r
            states[3:] = v
            errors = [error]
        else:
            states = np.empty((6, jd_f.size), dtype=np.float64)
            errors, r, v = satellite.sgp4_array(jd0, jd_f)
            states[:3, ...] = r.T
            states[3:, ...] = v.T

        for ind, err in enumerate(errors):
            if err != 0:
                logger.error(f"SGP4:propagate:step-{ind}:{SGP4_ERRORS[err]}")

        states *= 1e3  # km to m and km/s to m/s

        return states

    def TEME_to_TLE_OPTIM(
        self,
        cart: NDArray_6xN | NDArray_6,
        epoch: Time,
        t: NDArray_N | None = None,
        B: float = 0.0,
        tol: float = 1e-8,
        tol_v: float = 1e-9,
    ) -> NDArray_6:
        """Convert osculating orbital elements in TEME
        to mean elements used in two line element sets (TLE's).
        """
        logger.debug("SGP4:TEME_to_TLE_OPTIM")
        grav_model = getattr(sgp4.earth_gravity, self.settings.gravity_model.lower())

        if len(cart.shape) == 1:
            cart.shape = (cart.size, 1)

        if cart.shape[1] > 1 and t is None:
            raise ValueError(
                'Cannot convert TEME sampling to TLE without sample times "state_sample_times"'
            )
        elif t is None:
            t = np.array([0.0])

        t_min = np.argmin(np.abs(t))
        if t[t_min] > 1e-6:
            raise ValueError(
                "There is no sampling point at the epoch (t=0) to use as initial guess..."
            )

        init_elements = _cart2sgp4_elems(cart[:, t_min], grav_model=grav_model, degrees=False)

        t = epoch + TimeDelta(t, format="sec")

        def find_mean_elems(mean_elements):
            # Mean elements and osculating state
            state_osc = self.propagate_mean_elements(
                t.jd1,
                t.jd2,
                mean_elements,
                epoch.mjd - self.sgp4_mjd0,
                B=B,
            )

            d = cart - state_osc
            return np.mean(np.linalg.norm(d, axis=0))

        dx_std = np.array(self.settings.teme_to_tle_minimize_start_stds)
        samps = self.settings.teme_to_tle_minimize_start_samples
        bounds = self.settings.teme_to_tle_minimize_bounds

        opt_res = None
        for j in range(samps):
            _init_elements = init_elements.copy()
            if j > 0:
                _init_elements += np.random.randn(6) * dx_std

            for mni in range(6):
                if _init_elements[mni] < bounds[mni][0]:
                    _init_elements[mni] = bounds[mni][0]
                elif _init_elements[mni] > bounds[mni][1]:
                    _init_elements[mni] = bounds[mni][1]

            _opt_res = scipy.optimize.minimize(
                find_mean_elems,
                _init_elements,
                method="Nelder-Mead",
                bounds=bounds,
                options={
                    "fatol": np.sqrt(tol**2 + tol_v**2),
                },
            )
            if j > 0:
                if _opt_res.fun < opt_res.fun:  # type: ignore
                    opt_res = _opt_res
            else:
                opt_res = _opt_res

        mean_elements = opt_res.x  # type: ignore

        logger.debug("SGP4:TEME_to_TLE_OPTIM:completed")

        return mean_elements

    def TEME_to_TLE(
        self,
        cart: NDArray_6xN | NDArray_6,
        epoch: Time,
        t: NDArray_N | None = None,
        B: float = 0.0,
        tol: float = 1e-5,
        tol_v: float = 1e-7,
    ) -> NDArray_6:
        """Convert osculating orbital elements in TEME
        to mean elements used in two line element sets (TLE's).

        Parameters
        ----------
        cart
            Osculating State (position and velocity) vector in m and m/s,
            TEME frame.
        tol
            Wanted precision in position of mean element conversion in m.
        tol_v
            Wanted precision in velocity mean element conversion in m/s.

        Notes
        ----------
        mean elements of:
        - semi major axis (km)
        - orbital eccentricity
        - orbital inclination (radians)
        - right ascension of ascending node (radians)
        - argument of perigee (radians)
        - mean anomaly (radians)

        """
        logger.debug("SGP4:TEME_to_TLE")
        mean_elements = None
        grav_model = getattr(sgp4.earth_gravity, self.settings.gravity_model.lower())

        if len(cart.shape) > 1:
            if cart.size > 6:
                mean_elements = self.TEME_to_TLE_OPTIM(
                    cart,
                    epoch=epoch,
                    t=t,
                    B=B,
                    tol=tol,
                    tol_v=tol_v,
                )

                logger.debug("SGP4:TEME_to_TLE:completed")

                return mean_elements
            else:
                cart.shape = (cart.size,)

        state_mean = cart.copy()
        iter_max = self.settings.teme_to_tle_max_iter
        dr = 0
        dv = 0
        # Iterative determination of mean elements
        for it in range(iter_max):
            # Mean elements and osculating state
            mean_elements = _cart2sgp4_elems(state_mean, grav_model, degrees=False)

            if it > 0 and mean_elements[1] > 1:
                # Assumptions of osculation within slope not working
                # go to general minimization algorithms
                mean_elements = self.TEME_to_TLE_OPTIM(
                    cart,
                    epoch=epoch,
                    B=B,
                    tol=tol,
                    tol_v=tol_v,
                )
                break

            state_osc = self.propagate_mean_elements(
                epoch.jd1,
                epoch.jd2,
                mean_elements,
                epoch.mjd - self.sgp4_mjd0,
                B=B,
            )

            # Correction of mean state vector
            d = cart - state_osc
            state_mean += d
            if it > 0:
                dr_old = dr
                dv_old = dv

            dr = np.linalg.norm(d[:3, ...], axis=0)  # Position change
            dv = np.linalg.norm(d[3:, ...], axis=0)  # Velocity change

            if it > 0:
                if dr_old < dr or dv_old < dv:
                    # Assumptions of osculation within slope not working
                    # go to general minimization algorithms
                    mean_elements = self.TEME_to_TLE_OPTIM(
                        cart,
                        epoch=epoch,
                        B=B,
                        tol=tol,
                        tol_v=tol_v,
                    )
                    break

            if dr < tol and dv < tol_v:  # Iterate until position changes by less than eps
                break
            if it == iter_max - 1:
                # Iterative method not working, go to general minimization algorithms
                mean_elements = self.TEME_to_TLE_OPTIM(
                    cart,
                    epoch=epoch,
                    B=B,
                    tol=tol,
                    tol_v=tol_v,
                )

        logger.debug("SGP4:TEME_to_TLE:completed")

        return mean_elements  # type: ignore

    def space_object_to_mean_elements(
        self,
        space_object: SpaceObject,
    ) -> NDArray_6:
        assert space_object.state.num <= 1
        samples = self.settings.kepler_samples
        max_ang = 360.0 if space_object.state.degrees else 2 * np.pi
        max_ang *= self.settings.kepler_extent

        _orb = space_object.state.copy()
        _orb.allocate(samples)
        _orb._kep[()] = space_object.state._kep[()]
        _orb._kep[5, :] = np.linspace(-max_ang/2, max_ang/2, num=samples, endpoint=False)
        _orb.calculate_cartesian()
        t_vec = _orb.mean_anomaly / _orb.mean_motion

        return self.TEME_to_TLE_OPTIM(
            cart=_orb._cart,
            epoch=space_object.epoch,
            t=t_vec,
            B=get_B(space_object.properties),
            tol=self.settings.tol,
            tol_v=self.settings.tol_v,
        )



"""
Functions for core functionalities of this subpackage

- Intended to be imported as a whole module when consuming
"""

# TODO: this module can be moved to top level as helper funcs of sorts pkg?

from dataclasses import dataclass
import typing as t
import pickle
from pathlib import Path
import numpy as np
import numpy.typing as npt
from tqdm import tqdm
from astropy.time import Time, TimeDelta
from sorts.types import Datetime64_us, EcefStates, Float64_as_sec, Datetime_Like, StateType
from sorts.utils import to_datetime64_us
from sorts.radar import Station
from sorts.space_object import SpaceObject
from .types import Passage
from sorts.interpolation import Interpolator
from sorts.propagator import Propagator
from sorts.population import Population


@dataclass
class InterpolatedPropagation:
    # todo: investigate if we can just sidestep most of the `datetime64` and just use `Time`?
    times: npt.NDArray[Datetime64_us]
    states: EcefStates
    interpolator: Interpolator
    epoch: Datetime64_us

    @classmethod
    def from_space_objects(
        cls,
        space_objects: t.Sequence[SpaceObject] | Population,
        propagator: Propagator,
        interpolator_class: t.Type[Interpolator],
        start_time: Time,
        end_time: Time,
        time_step: float,
        progress: bool = False,
    ) -> list[t.Self]:
        prop_interps = []
        if progress:
            pbar = tqdm(desc="Propagating", total=len(space_objects))
        for spobj in space_objects:
            pint = cls.from_space_object(
                spobj,
                propagator,
                interpolator_class,
                start_time,
                end_time,
                time_step,
            )
            prop_interps.append(pint)
            if progress:
                pbar.update(1)
        if progress:
            pbar.close()
        return prop_interps

    @classmethod
    def from_space_object(
        cls,
        space_objects: SpaceObject,
        propagator: Propagator,
        interpolator_class: t.Type[Interpolator],
        start_time: Time,
        end_time: Time,
        time_step: float,
    ) -> t.Self:
        dt = (end_time - start_time).sec
        t0 = (start_time - space_objects.epoch).sec
        t_obj = np.arange(t0, t0 + dt, time_step, dtype=np.float64)
        itrs_states = propagator.propagate(space_objects, t_obj)

        interp = interpolator_class(itrs_states, t_obj)
        pint = cls(
            times=(space_objects.epoch + TimeDelta(t_obj, format="sec")).datetime64,
            states=itrs_states,
            interpolator=interp,
            epoch=space_objects.epoch.datetime64,
        )
        return pint

    @property
    def relative_seconds(self):
        return (self.times - self.epoch) / np.timedelta64(1, "s")


def find_simultaneous_passages(
    dt: npt.NDArray[Float64_as_sec],
    space_object: SpaceObject,
    states: EcefStates,
    tx_station: Station,
    rx_stations: t.Sequence[Station],
    epoch: Datetime_Like,
    fov_kw=None,
) -> list[Passage]:
    """
    Finds all passages that are simultaneously inside all tx, rx stations' FOV.
    """
    # NOTE: based on the `find_passes` func in `src/sorts/passes.py`

    epoch = to_datetime64_us(epoch)

    passages: list[Passage] = []
    if fov_kw is None:
        fov_kw = {}

    enu = []
    check = np.full((len(dt),), True, dtype=bool)
    for station in [tx_station, *rx_stations]:
        enu_st = station.enu(states)
        enu.append(enu_st)

        check_st = station.field_of_view(states, **fov_kw)
        check = np.logical_and(check, check_st)

    inds = np.where(check)[0]

    if len(inds) == 0:
        return passages

    dind = np.diff(inds)
    splits = np.where(dind > 1)[0]

    splits = np.insert(splits, 0, -1)
    splits = np.insert(splits, len(splits), len(inds) - 1)
    splits += 1
    for si in range(len(splits) - 1):
        ps_inds = inds[splits[si] : splits[si + 1]]
        if len(ps_inds) == 0:
            continue

        start_time: Datetime64_us = (
            t.cast(np.timedelta64, (dt[ps_inds[0]] * 1e6).astype("timedelta64[us]")) + epoch
        )

        end_time: Datetime64_us = (
            t.cast(np.timedelta64, (dt[ps_inds[-1]] * 1e6).astype("timedelta64[us]")) + epoch
        )

        time_range = (start_time, end_time)
        passages.append(
            Passage(
                space_object=space_object,
                tx_station=tx_station,
                rx_stations=list(rx_stations),
                epoch=epoch,
                time_range=time_range,
            )
        )

    return passages


def find_passages(
    dt: npt.NDArray[Float64_as_sec],
    space_object: SpaceObject,
    states: EcefStates,
    tx_station: Station,
    rx_station: Station,
    epoch: Datetime_Like,
    fov_kw=None,
) -> list[Passage]:
    """
    Finds all passages that are simultaneously inside a tx-rx station pair's FOV.
    """
    passages = find_simultaneous_passages(
        dt=dt,
        space_object=space_object,
        states=states,
        tx_station=tx_station,
        rx_stations=[rx_station],
        epoch=epoch,
        fov_kw=fov_kw,
    )
    return passages


def duplicate_and_perturbate_space_object(
    space_object: SpaceObject,
    propagator: Propagator,
    interpolator_class: t.Type[Interpolator],
    start_time: Time,
    end_time: Time,
    time_step: float,
    perturbation_format: StateType = "cartesian",
    pert_val: tuple[float, float, float, float, float, float] = (
        1e-3, 1e-3, 1e-3, 1e-5, 1e-5, 1e-5  # fmt: skip
    ),
) -> list[tuple[SpaceObject, InterpolatedPropagation]]:
    """TODO: detail structure"""
    # duplicate list items
    perturbed_objects = []

    # perturbate all state variables and leave one original
    # i.e. len 7, [(true_spobj_list, true_prop list), (pert_spobj_prop_list, ...) ...x6]
    for idx in range(7):
        # the original spobj are left intact, the rest are copied and perturbed
        if idx == 0:
            new_obj = space_object
        else:
            new_obj = space_object.copy()

            if perturbation_format == "kepler":
                new_obj.state._kep[idx - 1, 0] += pert_val[idx - 1]
                new_obj.state.calculate_cartesian()
            elif perturbation_format == "cartesian":
                new_obj.state._cart[idx - 1, 0] += pert_val[idx - 1]
                new_obj.state.calculate_kepler()

        prop_interp = InterpolatedPropagation.from_space_object(
            space_objects=new_obj,
            propagator=propagator,
            interpolator_class=interpolator_class,
            start_time=start_time,
            end_time=end_time,
            time_step=time_step,
        )
        perturbed_objects.append((new_obj, prop_interp, ))
    return perturbed_objects


def ensure_directory_exist(dpath: str | Path):
    """Check if the directory exist and create it if not"""

    dpath = Path(dpath)

    if not dpath.exists():
        dpath.mkdir(parents=True)
    assert dpath.exists()
    assert dpath.is_dir()


def safe_pickle(obj, fpath: str | Path):
    """
    Use pickle to save an object to the specified file path, with a few extra steps to make the write operation safer:
    - The output directory will be created if not exists
    - We write to an tmp file first then rename that file, as a simple way to reduce risk of corrupted files
    """

    fpath = Path(fpath)

    ensure_directory_exist(fpath.parent)

    fpath_tmp = fpath.with_suffix(fpath.suffix + ".tmp")
    with open(fpath_tmp, "wb") as f:
        pickle.dump(obj, f)
    fpath_tmp.rename(fpath)



#!/usr/bin/env python

"""Wrapper for the Orekit propagator

**Links:**
    * `orekit <https://www.orekit.org/>`_
    * `orekit python <https://gitlab.orekit.org/orekit-labs/python-wrapper>`_
    * `orekit python guide <https://gitlab.orekit.org/orekit-labs/python-wrapper/wikis/Manual-Installation-of-Python-Wrapper>`_
    * `Hipparchus <https://www.hipparchus.org/>`_
    * `orekit 9.3 api <https://www.orekit.org/static/apidocs/index.html>`_
    * `JCC <https://pypi.org/project/JCC/>`_


**Example usage:**

Simple propagation showing time difference due to loading of model data.

.. code-block:: python

    from sorts.propagator import Orekit

"""

# Python standard import
import copy, pathlib, logging
import urllib.request

# Third party import
import numpy as np
import scipy
import scipy.constants
import orekit
import pyorb

try:
    from tqdm import tqdm
except ImportError:
    tqdm = None

# Local import
from .base import Propagator
from .. import dates as dates
from .. import frames

from orekit.pyhelpers import setup_orekit_curdir
from org.orekit.propagation.numerical import NumericalPropagator
from org.orekit.propagation import SpacecraftState
from org.orekit.frames import FramesFactory
from org.orekit.time import TimeScalesFactory, AbsoluteDate
from org.orekit.orbits import KeplerianOrbit, EquinoctialOrbit, CartesianOrbit, PositionAngle
from org.orekit.utils import Constants, IERSConventions, PVCoordinates
from org.orekit.bodies import CelestialBodyFactory, OneAxisEllipsoid
from org.orekit.python import PythonOrekitStepHandler
import org.orekit.models.earth.atmosphere as orekit_atm
import org

from orekit import JArray_double

logger = logging.getLogger(__name__)

JAVA_VM = False


def init_vm():
    global JAVA_VM
    if not JAVA_VM:
        orekit.initVM()
        JAVA_VM = True


def npdt2absdate(dt, utc):
    """
    Converts a numpy datetime64 value to an orekit AbsoluteDate
    """

    year, month, day, hour, minutes, seconds, microsecond = dates.npdt_to_date(dt)
    return AbsoluteDate(
        int(year),
        int(month),
        int(day),
        int(hour),
        int(minutes),
        float(seconds + microsecond * 1e-6),
        utc,
    )


def mjd2absdate(mjd, utc):
    """
    Converts a Modified Julian Date value to an orekit AbsoluteDate
    """

    return npdt2absdate(dates.mjd_to_npdt(mjd), utc)


class Orekit(Propagator):
    """Propagator class implementing the Orekit propagator.

    #TODO: Update docs according to new "settings dict" method.
    #TODO: Valdiate Orekit v10 compatibility

    :ivar list solarsystem_perturbers: List of strings of names of objects in the solarsystem that should be used for third body perturbation calculations. All objects listed at `CelestialBodyFactory <https://www.orekit.org/static/apidocs/org/orekit/bodies/CelestialBodyFactory.html>`_ are available.
    :ivar str in_frame: String identifying the input frame to be used. All frames listed at `FramesFactory <https://www.orekit.org/static/apidocs/org/orekit/frames/FramesFactory.html>`_ are available.
    :ivar str out_frame: String identifying the output frame to be used. All frames listed at `FramesFactory <https://www.orekit.org/static/apidocs/org/orekit/frames/FramesFactory.html>`_ are available.
    :ivar bool drag_force: Should drag force be included in propagation.
    :ivar bool radiation_pressure: Should radiation pressure force be included in propagation.
    :ivar bool frame_tidal_effects: Should coordinate frames include Tidal effects.
    :ivar str integrator: String representing the numerical integrator from the Hipparchus package to use. Any integrator listed at `Hipparchus nonstiff ode <https://www.hipparchus.org/apidocs/org/hipparchus/ode/nonstiff/package-summary.html>`_ is available.
    :ivar float minStep: Minimum time step allowed in the numerical orbit propagation given in seconds.
    :ivar float maxStep: Maximum time step allowed in the numerical orbit propagation given in seconds.
    :ivar float position_tolerance: Position tolerance in numerical orbit propagation errors given in meters.
    :ivar str earth_gravity: Gravitation model to use for calculating central acceleration force. Currently avalible options are `'HolmesFeatherstone'` and `'Newtonian'`. See `gravity <https://www.orekit.org/static/apidocs/org/orekit/forces/gravity/package-summary.html>`_.
    :ivar tuple gravity_order: A tuple of two integers for describing the order of spherical harmonics used in the `HolmesFeatherstoneAttractionModel <https://www.orekit.org/static/apidocs/org/orekit/forces/gravity/HolmesFeatherstoneAttractionModel.html>`_ model.
    :ivar str atmosphere: Atmosphere model used to calculate atmospheric drag. Currently available options are `'DTM2000'`. See `atmospheres <https://www.orekit.org/static/apidocs/org/orekit/forces/drag/atmosphere/package-summary.html>`_.
    :ivar str solar_activity: The model used for calculating solar activity and thereby the influx of solar radiation. Used in the atmospheric drag force model. Currently available options are `'Marshall'` for the `MarshallSolarActivityFutureEstimation <https://www.orekit.org/static/apidocs/org/orekit/forces/drag/atmosphere/data/MarshallSolarActivityFutureEstimation.html>`_.
    :ivar str constants_source: Controls which source for Earth constants to use. Currently avalible options are `'WGS84'` and `'JPL-IAU'`. See `constants <https://www.orekit.org/static/apidocs/org/orekit/utils/Constants.html>`_.
    :ivar float mu: Standard gravitational constant for the Earth.  Definition depend on the :class:`sorts.propagator.Orekit` constructor parameter :code:`constants_source`
    :ivar float R_earth: Radius of the Earth in m. Definition depend on the :class:`sorts.propagator.Orekit` constructor parameter :code:`constants_source`
    :ivar float f_earth: Flattening of the Earth (i.e. :math:`\\frac{a-b}{a}` ). Definition depend on the :class:`sorts.propagator.Orekit` constructor parameter :code:`constants_source`.
    :ivar float M_earth: Mass of the Earth in kg. Definition depend on the :class:`sorts.propagator.Orekit` constructor parameter :code:`constants_source`
    :ivar org.orekit.frames.Frame inputFrame: The orekit frame instance for the input frame.
    :ivar org.orekit.frames.Frame outputFrame: The orekit frame instance for the output frame.
    :ivar org.orekit.frames.Frame inertialFrame: The orekit frame instance for the inertial frame. If inputFrame is pseudo innertial this is the same as inputFrame.
    :ivar org.orekit.bodies.OneAxisEllipsoid body: The model ellipsoid representing the Earth.
    :ivar dict _forces: Dictionary of forces to include in the numerical integration. Contains instances of children of :class:`org.orekit.forces.AbstractForceModel`.
    :ivar list _tolerances: Contains the absolute and relative tolerances calculated by the `tolerances <https://www.orekit.org/static/apidocs/org/orekit/propagation/numerical/NumericalPropagator.html#tolerances(double,org.orekit.orbits.Orbit,org.orekit.orbits.OrbitType)>`_ function.
    :ivar org.orekit.propagation.numerical.NumericalPropagator propagator: The numerical propagator instance.
    :ivar org.orekit.models.earth.atmosphere.data.MarshallSolarActivityFutureEstimation.StrengthLevel SolarStrengthLevel: The strength of the solar activity. Options are 'AVRAGE', 'STRONG', 'WEAK'.

    The constructor creates a propagator instance with supplied options.

    :param list solarsystem_perturbers: List of strings of names of objects in the solarsystem that should be used for third body perturbation calculations. All objects listed at `CelestialBodyFactory <https://www.orekit.org/static/apidocs/org/orekit/bodies/CelestialBodyFactory.html>`_ are available.
    :param str in_frame: String identifying the input frame to be used. All frames listed at `FramesFactory <https://www.orekit.org/static/apidocs/org/orekit/frames/FramesFactory.html>`_ are available.
    :param str out_frame: String identifying the output frame to be used. All frames listed at `FramesFactory <https://www.orekit.org/static/apidocs/org/orekit/frames/FramesFactory.html>`_ are available.
    :param bool drag_force: Should drag force be included in propagation.
    :param bool radiation_pressure: Should radiation pressure force be included in propagation.
    :param bool frame_tidal_effects: Should coordinate frames include Tidal effects.
    :param str integrator: String representing the numerical integrator from the Hipparchus package to use. Any integrator listed at `Hipparchus nonstiff ode <https://www.hipparchus.org/apidocs/org/hipparchus/ode/nonstiff/package-summary.html>`_ is available.
    :param float min_step: Minimum time step allowed in the numerical orbit propagation given in seconds.
    :param float max_step: Maximum time step allowed in the numerical orbit propagation given in seconds.
    :param float position_tolerance: Position tolerance in numerical orbit propagation errors given in meters.
    :param str atmosphere: Atmosphere model used to calculate atmospheric drag. Currently available options are `'DTM2000'`. See `atmospheres <https://www.orekit.org/static/apidocs/org/orekit/forces/drag/atmosphere/package-summary.html>`_.
    :param str solar_activity: The model used for calculating solar activity and thereby the influx of solar radiation. Used in the atmospheric drag force model. Currently available options are `'Marshall'` for the `MarshallSolarActivityFutureEstimation <https://www.orekit.org/static/apidocs/org/orekit/forces/drag/atmosphere/data/MarshallSolarActivityFutureEstimation.html>`_.
    :param str constants_source: Controls which source for Earth constants to use. Currently avalible options are `'WGS84'` and `'JPL-IAU'`. See `constants <https://www.orekit.org/static/apidocs/org/orekit/utils/Constants.html>`_.
    :param str earth_gravity: Gravitation model to use for calculating central acceleration force. Currently avalible options are `'HolmesFeatherstone'` and `'Newtonian'`. See `gravity <https://www.orekit.org/static/apidocs/org/orekit/forces/gravity/package-summary.html>`_.
    :param tuple gravity_order: A tuple of two integers for describing the order of spherical harmonics used in the `HolmesFeatherstoneAttractionModel <https://www.orekit.org/static/apidocs/org/orekit/forces/gravity/HolmesFeatherstoneAttractionModel.html>`_ model.
    :param str solar_activity_strength: The strength of the solar activity. Options are 'AVRAGE', 'STRONG', 'WEAK'.
    """

    class OrekitVariableStep(PythonOrekitStepHandler):
        """Class for handling the steps."""

        def set_params(self, t, start_date, states_pointer, outputFrame, heartbeat):
            self.t = t
            self.start_date = start_date
            self.states_pointer = states_pointer
            self.outputFrame = outputFrame
            self.__heartbeat = heartbeat

        def init(self, s0, t):
            pass

        def handleStep(self, interpolator, isLast):
            state1 = interpolator.getCurrentState()
            state0 = interpolator.getPreviousState()

            t0 = state0.getDate().durationFrom(self.start_date)
            t1 = state1.getDate().durationFrom(self.start_date)

            t_filt = np.logical_and(np.abs(self.t) >= np.abs(t0), np.abs(self.t) <= np.abs(t1))

            for ti, t in zip(np.where(t_filt)[0], self.t[t_filt]):
                t_date = self.start_date.shiftedBy(float(t))

                _state = interpolator.getInterpolatedState(t_date)

                PVCoord = _state.getPVCoordinates(self.outputFrame)

                x_tmp = PVCoord.getPosition()
                v_tmp = PVCoord.getVelocity()

                self.states_pointer[0, ti] = x_tmp.getX()
                self.states_pointer[1, ti] = x_tmp.getY()
                self.states_pointer[2, ti] = x_tmp.getZ()
                self.states_pointer[3, ti] = v_tmp.getX()
                self.states_pointer[4, ti] = v_tmp.getY()
                self.states_pointer[5, ti] = v_tmp.getZ()

                if self.__heartbeat is not None:
                    self.__heartbeat(
                        float(t), self.states_pointer[:, ti], interpolator=interpolator
                    )

    DEFAULT_SETTINGS = copy.copy(Propagator.DEFAULT_SETTINGS)
    DEFAULT_SETTINGS.update(
        dict(
            in_frame="Orekit-EME",
            out_frame="Orekit-ITRF",
            frame_tidal_effects=False,
            integrator="DormandPrince853",
            min_step=0.001,
            max_step=120.0,
            position_tolerance=10.0,
            earth_gravity="HolmesFeatherstone",
            gravity_order=(10, 10),
            solarsystem_perturbers=["Moon", "Sun"],
            drag_force=True,
            atmosphere="DTM2000",
            radiation_pressure=True,
            solar_activity="Marshall",
            constants_source="WGS84",
            solar_activity_strength="WEAK",
        )
    )

    @staticmethod
    def download_quickstart_data(
        path, url=None, headers=None, timeout=10.0, block_size=2048, verbose=True
    ):
        if url is None:
            # Standard URL of the example data distributed on orekit.org
            url = "https://gitlab.orekit.org/orekit/orekit-data/-/archive/master/orekit-data-master.zip"
        if headers is None:
            headers = {}

        pbar_str = f'Downloading from {url[:15]}...{url.split("/")[-1]} [kb]'

        if path.parent.exists():
            if not path.parent.is_dir():
                raise FileExistsError(f"{path.parent} exists and is not a directory")
        else:
            print(f"Creating directory {path.parent} for orekit data master")
            path.parent.mkdir(parents=True)

        urlopener = urllib.request.build_opener()
        request = urllib.request.Request(url, headers=headers)
        with urlopener.open(request, timeout=timeout) as remote:
            info = remote.info()
            try:
                size = int(info["Content-Length"])
                size_kb = int(size // 1024)
            except (KeyError, ValueError, TypeError):
                size = None
                size_kb = None

            if verbose:
                pbar = tqdm(total=size_kb)

            with open(path, "wb") as f:
                try:
                    bytes_done = 0
                    block = remote.read(block_size)
                    while block:
                        f.write(block)
                        bytes_done += len(block)

                        if verbose:
                            pbar.set_description(pbar_str)
                            pbar.update(int(len(block) // 1024))

                        block = remote.read(block_size)

                    if size is not None and bytes_done != size:
                        raise Exception(
                            f"Content-Length {size} did not match download size {bytes_done}, aborting..."
                        )
                except BaseException:
                    if path.is_file():
                        try:
                            print("Attempting to remove failed partial download")
                            path.unlink()
                        except:
                            pass
                    raise

    def __init__(self, orekit_data, settings=None, **kwargs):
        super(Orekit, self).__init__(settings=settings, **kwargs)
        init_vm()

        logger.debug(f"sorts.propagator.Orekit:init")

        if isinstance(orekit_data, pathlib.Path):
            orekit_data = str(orekit_data)

        setup_orekit_curdir(filename=orekit_data)

        logger.debug(f"Orekit:init:orekit-data = {orekit_data}")

        self.utc = TimeScalesFactory.getUTC()

        self.__settings = dict()
        self.__settings["SolarStrengthLevel"] = getattr(
            orekit_atm.data.MarshallSolarActivityFutureEstimation.StrengthLevel,
            self.settings["solar_activity_strength"],
        )
        self._tolerances = None

        if self.settings["constants_source"] == "JPL-IAU":
            self.mu = Constants.JPL_SSD_EARTH_GM
            self.R_earth = Constants.IAU_2015_NOMINAL_EARTH_EQUATORIAL_RADIUS
            R_diff = (
                Constants.IAU_2015_NOMINAL_EARTH_EQUATORIAL_RADIUS
                - Constants.IAU_2015_NOMINAL_EARTH_POLAR_RADIUS
            )
            self.f_earth = R_diff / Constants.IAU_2015_NOMINAL_EARTH_POLAR_RADIUS
        else:
            self.mu = Constants.WGS84_EARTH_MU
            self.R_earth = Constants.WGS84_EARTH_EQUATORIAL_RADIUS
            self.f_earth = Constants.WGS84_EARTH_FLATTENING

        self.M_earth = self.mu / scipy.constants.G

        self.__params = None

        self._forces = {}

        if self.settings["radiation_pressure"]:
            self._forces["radiation_pressure"] = None
        if self.settings["drag_force"]:
            self._forces["drag_force"] = None

        if self.settings["earth_gravity"] == "HolmesFeatherstone":
            provider = (
                org.orekit.forces.gravity.potential.GravityFieldFactory.getNormalizedProvider(
                    self.settings["gravity_order"][0],
                    self.settings["gravity_order"][1],
                )
            )
            holmesFeatherstone = org.orekit.forces.gravity.HolmesFeatherstoneAttractionModel(
                FramesFactory.getITRF(IERSConventions.IERS_2010, True),
                provider,
            )
            self._forces["earth_gravity"] = holmesFeatherstone

        elif self.settings["earth_gravity"] == "Newtonian":
            Newtonian = org.orekit.forces.gravity.NewtonianAttraction(self.mu)
            self._forces["earth_gravity"] = Newtonian

        else:
            raise Exception(
                f'Supplied Earth gravity model "{self.settings["earth_gravity"]}" not recognized'
            )

        if self.settings["solarsystem_perturbers"] is not None:
            for body in self.settings["solarsystem_perturbers"]:
                body_template = getattr(CelestialBodyFactory, "get{}".format(body))
                body_instance = body_template()
                perturbation = org.orekit.forces.gravity.ThirdBodyAttraction(body_instance)

                self._forces["perturbation_{}".format(body)] = perturbation

        for key in self._forces:
            if self._forces[key] is not None:
                logger.debug(f"Orekit:init:_forces:{key} = {type(self._forces[key])}")
            else:
                logger.debug(f"Orekit:init:_forces:{key} = None")

    def __str__(self):
        ret = ""
        ret += "Orekit instance @ {}:".format(hash(self)) + "\n" + "-" * 25 + "\n"
        ret += "{:20s}: ".format("Integrator") + self.settings["integrator"] + "\n"
        ret += "{:20s}: ".format("Minimum step") + str(self.settings["min_step"]) + " s" + "\n"
        ret += "{:20s}: ".format("Maximum step") + str(self.settings["max_step"]) + " s" + "\n"
        ret += (
            "{:20s}: ".format("Position Tolerance")
            + str(self.settings["position_tolerance"])
            + " m"
            + "\n"
        )
        if self._tolerances is not None:
            ret += (
                "{:20s}: ".format("Absolute Tolerance")
                + str(JArray_double.cast_(self._tolerances[0]))
                + " m"
                + "\n"
            )
            ret += (
                "{:20s}: ".format("Relative Tolerance")
                + str(JArray_double.cast_(self._tolerances[1]))
                + " m"
                + "\n"
            )
        ret += "\n"
        ret += "{:20s}: ".format("Input frame") + self.settings["in_frame"] + "\n"
        ret += "{:20s}: ".format("Output frame") + self.settings["out_frame"] + "\n"
        ret += "{:20s}: ".format("Gravity model") + self.settings["earth_gravity"] + "\n"
        if self.settings["earth_gravity"] == "HolmesFeatherstone":
            ret += (
                "{:20s} - Harmonic expansion order {}".format("", self.settings["gravity_order"])
                + "\n"
            )
        ret += "{:20s}: ".format("Atmosphere model") + self.settings["atmosphere"] + "\n"
        ret += "{:20s}: ".format("Solar model") + self.settings["solar_activity"] + "\n"
        ret += "{:20s}: ".format("Constants") + self.settings["constants_source"] + "\n"
        ret += "Included forces:" + "\n"
        for key in self._forces:
            ret += " - {}".format(" ".join(key.split("_"))) + "\n"
        ret += "Third body perturbations:" + "\n"
        for body in self.settings["solarsystem_perturbers"]:
            ret += " - {:}".format(body) + "\n"

        return ret

    def _get_frame(self, name):
        """Uses a string to identify which coordinate frame to initialize from Orekit package.

        See `Orekit FramesFactory <https://www.orekit.org/static/apidocs/org/orekit/frames/FramesFactory.html>`_
        """

        if name == "EME":
            frame = FramesFactory.getEME2000()
        elif name == "EME2000":
            frame = FramesFactory.getEME2000()
        elif name == "GCRF":
            frame = FramesFactory.getGCRF()
        elif name == "CIRF":
            frame = FramesFactory.getCIRF(
                IERSConventions.IERS_2010, not self.settings["frame_tidal_effects"]
            )
        elif name == "ITRF":
            frame = FramesFactory.getITRF(
                IERSConventions.IERS_2010, not self.settings["frame_tidal_effects"]
            )
        elif name == "TIRF":
            frame = FramesFactory.getTIRF(
                IERSConventions.IERS_2010, not self.settings["frame_tidal_effects"]
            )
        elif name == "ITRFEquinox":
            frame = FramesFactory.getITRFEquinox(
                IERSConventions.IERS_2010, not self.settings["frame_tidal_effects"]
            )
        elif name == "TEME":
            frame = FramesFactory.getTEME()
        else:
            raise Exception('Frame "{}" not recognized'.format(name))

        return frame

    def _construct_propagator(self, initialOrbit):
        """
        Get the specified integrator from hipparchus package. List available at:
        `nonstiff ode <https://www.hipparchus.org/apidocs/org/hipparchus/ode/nonstiff/package-summary.html>`_

        Configure the integrator tolerances using the orbit.
        """

        self._tolerances = NumericalPropagator.tolerances(
            self.settings["position_tolerance"], initialOrbit, initialOrbit.getType()
        )

        integrator_constructor = getattr(
            org.hipparchus.ode.nonstiff,
            "{}Integrator".format(self.settings["integrator"]),
        )

        integrator = integrator_constructor(
            self.settings["min_step"],
            self.settings["max_step"],
            JArray_double.cast_(self._tolerances[0]),
            JArray_double.cast_(self._tolerances[1]),
        )

        propagator = NumericalPropagator(integrator)
        propagator.setOrbitType(initialOrbit.getType())

        self.propagator = propagator

    def _set_forces(self, A, cd, cr):
        """Using the spacecraft specific parameters, set the drag force and radiation pressure models.

        **See:**
            * `drag <https://www.orekit.org/static/apidocs/org/orekit/forces/drag/package-summary.html>`_
            * `radiation <https://www.orekit.org/static/apidocs/org/orekit/forces/radiation/package-summary.html>`_
        """

        logger.debug(f"Orekit:set_forces:A = {A}")
        logger.debug(f"Orekit:set_forces:cd = {cd}")
        logger.debug(f"Orekit:set_forces:cr = {cr}")

        __params = [A, cd, cr]

        re_calc = True
        if self.__params is None:
            re_calc = True
        else:
            if not np.allclose(np.array(__params, dtype=np.float64), self.__params, rtol=1e-3):
                re_calc = True

        logger.debug(f"Orekit:set_forces:re_calc = {re_calc}")

        self.atmosphere_instance = None
        self.spacecraft_drag_model = None

        if re_calc:
            self.__params = __params
            self.force_params = {"A": A, "C_D": cd, "C_R": cr}
            if self.settings["drag_force"]:
                if self.settings["solar_activity"] == "Marshall":
                    self._msafe = orekit_atm.data.MarshallSolarActivityFutureEstimation(
                        orekit_atm.data.MarshallSolarActivityFutureEstimation.DEFAULT_SUPPORTED_NAMES,
                        self.__settings["SolarStrengthLevel"],
                    )
                    self._manager = org.orekit.data.DataProvidersManager.getInstance()
                    self._manager.feed(self._msafe.getSupportedNames(), self._msafe)

                    if self.settings["atmosphere"] == "DTM2000":
                        self.atmosphere_instance = orekit_atm.DTM2000(
                            self._msafe,
                            CelestialBodyFactory.getSun(),
                            self.body,
                            self.utc,
                        )
                    else:
                        raise Exception("Atmosphere model not recognized")

                    self._forces["drag_force"] = self.GetIsotropicDragForce(A, cd)
                else:
                    raise Exception("Solar activity model not recognized")

            if self.settings["radiation_pressure"]:
                radiation_pressure_model = org.orekit.forces.radiation.SolarRadiationPressure(
                    CelestialBodyFactory.getSun(),
                    self.body.getEquatorialRadius(),
                    org.orekit.forces.radiation.IsotropicRadiationSingleCoefficient(
                        float(A), float(cr)
                    ),
                )

                self._forces["radiation_pressure"] = radiation_pressure_model

            # self.propagator.removeForceModels()

            self.UpdateForces()

    def UpdateForces(self):
        for force_name, force in self._forces.items():
            self.propagator.addForceModel(force)

    def GetIsotropicDragForce(self, A, cd):
        self.spacecraft_drag_model = org.orekit.forces.drag.IsotropicDrag(float(A), float(cd))
        force_ = org.orekit.forces.drag.DragForce(
            self.atmosphere_instance,
            self.spacecraft_drag_model,
        )
        return force_

    def propagate(self, t, state0, epoch, **kwargs):
        """
        **Implementation:**

        Keyword arguments are:

            * float A: Area in m^2
            * float C_D: Drag coefficient
            * float C_R: Radiation pressure coefficient
            * float m: Mass of object in kg
        """

        if self.settings["in_frame"].startswith("Orekit-"):
            self.inputFrame = self._get_frame(self.settings["in_frame"].replace("Orekit-", ""))
            orekit_in_frame = True
        else:
            self.inputFrame = self._get_frame("GCRF")
            orekit_in_frame = False

        if self.settings["out_frame"].startswith("Orekit-"):
            self.outputFrame = self._get_frame(self.settings["out_frame"].replace("Orekit-", ""))
            orekit_out_frame = True
        else:
            self.outputFrame = self._get_frame("GCRF")
            orekit_out_frame = False

        if self.inputFrame.isPseudoInertial():
            self.inertialFrame = self.inputFrame
        else:
            self.inertialFrame = FramesFactory.getEME2000()

        self.body = OneAxisEllipsoid(self.R_earth, self.f_earth, self.outputFrame)

        if isinstance(state0, pyorb.Orbit):
            state0_cart = np.squeeze(state0.cartesian)
        else:
            state0_cart = state0

        if self.settings["radiation_pressure"]:
            if "C_R" not in kwargs:
                raise Exception('Radiation pressure force enabled but no coefficient "C_R" given')
        else:
            kwargs["C_R"] = 1.0

        if self.settings["drag_force"]:
            if "C_D" not in kwargs:
                raise Exception('Drag force enabled but no drag coefficient "C_D" given')
            if "A" not in kwargs:
                raise Exception('Drag force enabled but no area "A" given')
        else:
            if "C_D" not in kwargs:
                kwargs["C_D"] = 2.3
            if "A" not in kwargs:
                kwargs["A"] = 1.0

        if "m" not in kwargs:
            kwargs["m"] = 0.0

        t, epoch = self.convert_time(t, epoch)
        times = epoch + t

        mjd0 = epoch.mjd
        t = t.sec
        if not isinstance(t, np.ndarray):
            t = np.array([t])

        logger.debug(f"Orekit:propagate:len(t) = {len(t)}")

        if not orekit_in_frame:
            state0_cart = frames.convert(
                epoch,
                state0_cart,
                in_frame=self.settings["in_frame"],
                out_frame="GCRS",
            )

        initialDate = mjd2absdate(mjd0, self.utc)

        pos = org.hipparchus.geometry.euclidean.threed.Vector3D(
            float(state0_cart[0]),
            float(state0_cart[1]),
            float(state0_cart[2]),
        )
        vel = org.hipparchus.geometry.euclidean.threed.Vector3D(
            float(state0_cart[3]),
            float(state0_cart[4]),
            float(state0_cart[5]),
        )
        PV_state = PVCoordinates(pos, vel)

        if not self.inputFrame.isPseudoInertial():
            transform = self.inputFrame.getTransformTo(self.inertialFrame, initialDate)
            PV_state = transform.transformPVCoordinates(PV_state)

        initialOrbit = CartesianOrbit(
            PV_state,
            self.inertialFrame,
            initialDate,
            self.mu + float(scipy.constants.G * kwargs["m"]),
        )

        self._construct_propagator(initialOrbit)
        self._set_forces(kwargs["A"], kwargs["C_D"], kwargs["C_R"])

        initialState = SpacecraftState(initialOrbit)

        self.propagator.setInitialState(initialState)

        tb_inds = t < 0.0
        t_back = t[tb_inds]

        tf_indst = t >= 0.0
        t_forward = t[tf_indst]

        if len(t_forward) == 1:
            if np.any(t_forward == 0.0):
                t_back = t
                t_forward = []
                tb_inds = t <= 0

        state = np.empty((6, len(t)), dtype=np.float64)
        step_handler = Orekit.OrekitVariableStep()

        if self.settings["heartbeat"]:
            __heartbeat = self.heartbeat
        else:
            __heartbeat = None

        if len(t_back) > 0:
            _t = t_back
            _t_order = np.argsort(np.abs(_t))
            _t_res = np.argsort(_t_order)
            _t = _t[_t_order]
            _state = np.empty((6, len(_t)), dtype=np.float64)

            step_handler.set_params(
                _t,
                initialDate,
                _state,
                self.outputFrame,
                __heartbeat,
            )

            self.propagator.setMasterMode(step_handler)

            t_next = initialDate.shiftedBy(float(_t[-1]))
            self.propagator.propagate(t_next)

            # now _state is full and in the order of _t
            state[:, tb_inds] = _state[:, _t_res]

        if len(t_forward) > 0:
            _t = t_forward
            _t_order = np.argsort(np.abs(_t))
            _t_res = np.argsort(_t_order)
            _t = _t[_t_order]
            _state = np.empty((6, len(_t)), dtype=np.float64)
            step_handler.set_params(
                _t,
                initialDate,
                _state,
                self.outputFrame,
                __heartbeat,
            )

            self.propagator.setMasterMode(step_handler)

            t_next = initialDate.shiftedBy(float(_t[-1]))
            self.propagator.propagate(t_next)

            # now _state is full and in the order of _t
            state[:, tf_indst] = _state[:, _t_res]

        if not orekit_out_frame:
            state = frames.convert(
                times,
                state,
                in_frame="GCRS",
                out_frame=self.settings["out_frame"],
            )

        logger.debug(f"Orekit:propagate:completed")

        return state



#!/usr/bin/env python

"""Minimoon population

"""
import pathlib

import numpy as np
from astropy.time import Time
import pyorb

from .population import Population


def NESCv9_minimoons(
    fname,
    albedo,
    propagator=None,
    propagator_options={},
    propagator_args={},
):
    """
    input file data format:

    1. synthetic orbit designation
    2. orbital element type (KEP for heliocentric Keplerian)
    3. semimajor axis (au)
    4. eccentricity
    5. inclination (deg)
    6. longitude of ascending node (deg)
    7. argument of perihelion (deg)
    8. mean anomaly (deg)
    9. H-magnitude (filler value)
    10. epoch for which the input orbit is valid (modified Julian date)
    11. index (1)
    12. number of parameters (6)
    13. MOID (filler value)
    14. Code with which created (OPENORB)
    """

    data = np.genfromtxt(fname)
    pop = Population(
        fields=["oid", "a", "e", "i", "aop", "raan", "mu0", "mjd0", "d"],
        dtypes=["int"] + ["float64"] * 8,
        space_object_fields=["d"],
        state_fields=["a", "e", "i", "aop", "raan", "mu0"],
        epoch_field={"field": "mjd0", "format": "mjd", "scale": "utc"},
        propagator=propagator,
        propagator_options=propagator_options,
        propagator_args=propagator_args,
    )

    pop.allocate(data.shape[0])

    pop["oid"] = np.arange(len(pop))
    pop["a"] = data[:, 2] * pyorb.AU
    pop["e"] = data[:, 3]
    pop["i"] = data[:, 4]
    pop["raan"] = data[:, 5]
    pop["aop"] = data[:, 6]
    pop["mu0"] = data[:, 7]
    pop["mjd0"] = data[:, 9]
    pop["d"] = 10.0 ** (3.1236 - 0.5 * np.log10(albedo) - 0.2 * data[:, 8])

    return pop



#!/usr/bin/env python

"""

"""

import numpy as np

from .scan import Scan


class RandomUniform(Scan):
    """Uniform randomly distributed points in the FOV."""

    def __init__(self, min_elevation=30.0, dwell=0.1, cycle_num=10000):
        super().__init__(coordinates="enu")
        self._dwell = dwell
        self.num = cycle_num
        self.min_elevation = min_elevation

        min_z = np.sin(np.radians(min_elevation))

        theta = 2 * np.pi * np.random.rand(self.num)
        phi = np.arccos(np.random.rand(self.num) * (1 - min_z) + min_z)

        k = np.empty((3, self.num), dtype=np.float64)

        k[0, :] = np.cos(theta) * np.sin(phi)
        k[1, :] = np.sin(theta) * np.sin(phi)
        k[2, :] = np.cos(phi)

        self.pointings = k

    def dwell(self, t=None):
        if t is None:
            return self._dwell
        else:
            if isinstance(t, float) or isinstance(t, int):
                return self._dwell
            else:
                return np.ones(t.shape, dtype=t.dtype) * self._dwell

    def min_dwell(self):
        return self._dwell

    def cycle(self):
        return self.num * self._dwell

    def pointing(self, t):
        ind = (np.mod(t / self.cycle(), 1) * self.num).astype(np.int64)
        return self.pointings[:, ind]



import typing as t
import numpy.typing as npt
import bokeh.plotting as bp
import bokeh.models as bokeh_models
from sorts.types import Float64_as_deg
from sorts.utils import assert_class_attributes_equal_to


AzelSkyplotColumnKey = t.Literal["azimuth", "elevation", "adj_azimuth", "adj_elevation"]


class _K_AzelSkyplotColumnKey:
    """Internal helper class for accessing string keys consistently"""

    azimuth: t.Final = "azimuth"
    elevation: t.Final = "elevation"
    adj_azimuth: t.Final = "adj_azimuth"
    adj_elevation: t.Final = "adj_elevation"


assert_class_attributes_equal_to(_K_AzelSkyplotColumnKey, t.get_args(AzelSkyplotColumnKey))


def _azel_skyplot_from_cds(source: bokeh_models.ColumnarDataSource):
    """An internal ver of `azel_skyplot` that takes a bokeh `ColumnDataSource`."""

    _K = _K_AzelSkyplotColumnKey

    # make a plot and set the pixel aspect ratio to equal to the data aspect ratio
    # (i.e. a circle in data will be a circle on screen)
    plot = bp.figure(match_aspect=True)
    # customize axis
    plot.xaxis.fixed_location = 0
    plot.yaxis.fixed_location = 0
    plot.xaxis.ticker = [30, 60, 90]
    plot.xaxis.major_label_overrides = {30: "60", 60: "30", 90: "0"}
    plot.yaxis.ticker = []
    plot.xaxis.axis_line_alpha = 0  # alternatively, `plot.xaxis.axis_line_color = "lightgray"`
    plot.yaxis.axis_line_alpha = 0  # alternatively, `plot.yaxis.axis_line_color = "lightgray"`

    # disable builtin grid, which is rectangular, we will draw a custom polar grid
    plot.xgrid.visible = False
    plot.ygrid.visible = False

    # add a bit more padding to the plotting region, default is 0.1 (in ratio)
    plot.x_range.range_padding = 0.15  # type: ignore
    plot.y_range.range_padding = 0.15  # type: ignore

    # disable per axis zoom when hovered on an axis
    wheelZoomTool = next(
        (t for t in plot.toolbar.tools if isinstance(t, bokeh_models.WheelZoomTool))
    )
    wheelZoomTool.zoom_on_axis = False

    # draw custom grid
    ray_angles = [0, 30, 60, 90, 120, 150, 180, 210, 240, 270, 300, 330]
    plot.ray(
        x=[0 for _ in range(len(ray_angles))],
        y=[0 for _ in range(len(ray_angles))],
        length=90,
        angle=ray_angles,
        angle_units="deg",
        color="lightgray",
        line_width=1,
    )
    ring_sizes = [0, 30, 60, 90]  # in deg
    plot.circle(
        y=[0 for _ in range(len(ring_sizes))],
        x=[0 for _ in range(len(ring_sizes))],
        radius=ring_sizes,
        color="lightgray",
        fill_alpha=0,
        line_width=1,
    )

    # add custom labels
    for x, y, text, angle in [
        (0, 100, "N, 0°", 0),
        (100, 0, "E, 90°", -90),
        (0, -100, "S, 180°", 0),
        (-100, 0, "W, -90°", 90),
    ]:
        plot.add_layout(
            bokeh_models.Label(
                x=x,
                y=y,
                anchor="center",  # type: ignore
                text=text,
                angle=angle,
                angle_units="deg",
            )
        )

    tf = bokeh_models.PolarTransform(
        angle=_K.adj_azimuth,
        radius=_K.adj_elevation,
        angle_units="deg",  # type: ignore
        direction="clock",
    )

    scatter = plot.scatter(
        x=tf.x,  # type: ignore
        y=tf.y,  # type: ignore
        source=source,
    )

    # add custom tooltips, only for the scatter plot/glyphs
    plot.add_tools(
        bokeh_models.HoverTool(
            renderers=[scatter],
            tooltips=[
                ("index", "$index"),
                ("data (az, el)", f"(@{_K.azimuth}, @{_K.elevation})"),
            ],
        )
    )

    return plot


def _azel_skyplot_cds_cols(
    azimuths: npt.NDArray[Float64_as_deg], elevations: npt.NDArray[Float64_as_deg]
):
    cols: dict[AzelSkyplotColumnKey, npt.NDArray] = {
        "azimuth": azimuths,
        "elevation": elevations,
        "adj_azimuth": azimuths - 90,
        "adj_elevation": 90 - elevations,
    }

    return cols


def azel_skyplot(azimuths: npt.NDArray[Float64_as_deg], elevations: npt.NDArray[Float64_as_deg]):
    """
    A plot similar to a skyplot.

    e.g. in the one in mathlab
    https://mathworks.com/help/satcom/ref/skyplot.html
    """

    cols = _azel_skyplot_cds_cols(azimuths, elevations)
    plot = _azel_skyplot_from_cds(source=bokeh_models.ColumnDataSource(t.cast(dict, cols)))

    return plot



import abc
from sorts import radar, scheduling


class ControllerBase(abc.ABC):
    @abc.abstractmethod
    def get_experiment_detail(self) -> scheduling.ExperimentDetail: ...

    @abc.abstractmethod
    def get_experiment_id_station_id_pairs_map(
        self,
    ) -> scheduling.ExperimentIdStationIdPairsMap: ...

    @abc.abstractmethod
    def get_station_map(self) -> dict[radar.StationId, radar.Station]: ...



"""Defines all the pre-configured radar instances

"""

# from . import eiscat_3d
# from . import tsdr
# from . import eiscat_uhf
# from . import eiscat_esr
from . import nostra
# from . import mock

from .radars import get_radar, list_radars



#!/usr/bin/env python

"""Constants used by SORTS from various sources.

"""

# Third party import
import scipy.constants


class WGS84:
    """World Geodetic System 1984 constants."""

    a = 6378.137 * 1e3
    """float: semi-major axis parameter in meters of the World Geodetic System 1984 (WGS84)
    """

    b = 6356.7523142 * 1e3
    """float: semi-minor axis parameter in meters of the World Geodetic System 1984 (WGS84)
    """

    esq = 6.69437999014 * 0.001
    """float: `esq` parameter in meters of the World Geodetic System 1984 (WGS84)
    """

    e1sq = 6.73949674228 * 0.001
    """float: `e1sq` parameter in meters of the World Geodetic System 1984 (WGS84)
    """

    f = 1 / 298.257223563
    """float: `f` parameter of the World Geodetic System 1984 (WGS84)
    """


R_earth = 6371.0088e3
"""float: Radius of the Earth using the International Union of Geodesy and Geophysics (IUGG) definition
"""


class WGS72:
    """World Geodetic System 1972 constants."""

    MU_earth = 398600.8 * 1e9
    """float: Standard gravitational parameter of the Earth using the WGS72 convention.
    """

    M_earth = MU_earth / scipy.constants.G
    """float: Mass of the Earth using the WGS72 convention.
    """



#!/usr/bin/env python

"""

"""

import numpy as np

from .scan import Scan


class Fence(Scan):
    """General fence scan."""

    def __init__(self, azimuth=0.0, min_elevation=30.0, dwell=0.2, num=20):
        super().__init__(coordinates="azelr")
        self._dwell = dwell
        self.num = num
        self.min_elevation = min_elevation
        self.azimuth = azimuth

        self._az = np.empty((num,), dtype=np.float64)
        self._el = np.linspace(min_elevation, 180 - min_elevation, num=num, dtype=np.float64)
        inds_ = self._el > 90.0
        self._az[inds_] = np.mod(self.azimuth + 180.0, 360.0)
        self._az[np.logical_not(inds_)] = self.azimuth

        self._el[inds_] = 180.0 - self._el[inds_]

    def dwell(self, t=None):
        if t is None:
            return self._dwell
        else:
            if isinstance(t, float) or isinstance(t, int):
                return self._dwell
            else:
                return np.ones(t.shape, dtype=t.dtype) * self._dwell

    def min_dwell(self):
        return self._dwell

    def cycle(self):
        return self.num * self._dwell

    def pointing(self, t):
        ind = (np.mod(t / self.cycle(), 1) * self.num).astype(np.int64)
        if isinstance(t, float) or isinstance(t, int):
            shape = (3,)
        else:
            shape = (3, len(t))

        azelr = np.empty(shape, dtype=np.float64)
        azelr[0, ...] = self._az[ind]
        azelr[1, ...] = self._el[ind]
        azelr[2, ...] = 1.0
        return azelr



#!/usr/bin/env python

"""Tracking plot functions

"""

# Python standard import


# Third party import
import numpy as np
import matplotlib.pyplot as plt


# Local import
from .. import frames


def az_el_to_xy(az, el):
    az, el = np.radians(az), np.radians(el)

    r = np.cos(el)
    x = r * np.sin(az)
    y = r * np.cos(az)
    return x, y


def local_passes(passes, **kwargs):
    enu_ind = kwargs.pop("station_ind", 0)
    fig = None
    ax = None

    for ind, ps in enumerate(passes):
        if "add_track" not in kwargs:
            if ind == 0:
                kwargs["add_track"] = False
            else:
                kwargs["add_track"] = True

        if isinstance(ps.enu, list):
            enu = ps.enu[enu_ind]
        else:
            enu = ps.enu

        azelr = frames.cart_to_sph(enu[:3, :], degrees=kwargs.setdefault("degrees", True))

        _fig, ax = local_tracking(azelr[0, :], azelr[1, :], **kwargs)
        if fig is None:
            fig = _fig

        if "ax" not in kwargs:
            kwargs["ax"] = ax

    return fig, ax


def local_tracking(
    azimuth,
    elevation,
    ax=None,
    t=None,
    track_line_style="-",
    track_marker=None,
    track_color=None,
    add_track=False,
    node_times=False,
    degrees=True,
    alpha=1,
):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        fig = None

    x0, y0 = az_el_to_xy(azimuth, elevation)
    ax.plot(x0, y0, c=track_color, ls=track_line_style, marker=track_marker, alpha=alpha)

    if not add_track:
        x, y = az_el_to_xy(np.linspace(0, 360, num=360), np.repeat(0.0, 360))
        ax.plot(x, y, color="green", alpha=0.5)

        x, y = az_el_to_xy(np.linspace(0, 360, num=360), np.repeat(30.0, 360))
        x[np.logical_and(x > 0, abs(y) < 0.1)] = np.nan
        ax.plot(x, y, color="black")
        ax.text(
            np.cos(np.pi * 30 / 180.0),
            0,
            r"$30\degree$",
            horizontalalignment="center",
            verticalalignment="center",
        )

        x, y = az_el_to_xy(np.linspace(0, 360, num=360), np.repeat(60.0, 360))
        x[np.logical_and(x > 0, abs(y) < 0.1)] = np.nan
        ax.plot(x, y, color="black")
        ax.text(
            np.cos(np.pi * 60 / 180.0),
            0,
            r"$60\degree$",
            horizontalalignment="center",
            verticalalignment="center",
        )

        x, y = az_el_to_xy(np.linspace(0, 360, num=360), np.repeat(80.0, 360))
        x[np.logical_and(x > 0, abs(y) < 0.1)] = np.nan
        ax.plot(x, y, color="black")
        ax.text(
            np.cos(np.pi * 80 / 180.0),
            0,
            r"$80\degree$",
            horizontalalignment="center",
            verticalalignment="center",
        )

    ax.plot(x0[0], y0[0], "o", alpha=alpha)
    ax.plot(x0[-1], y0[-1], "x", alpha=alpha)

    if t is None or not node_times:
        start = "Start"
        end = "End"
    else:
        start = t[0].to_value("isot", subfmt="date_hm")
        end = t[-1].to_value("isot", subfmt="date_hm")

    if node_times is not None:
        ax.text(x0[0], y0[0], start)
        ax.text(x0[-1], y0[-1], end)

    if not add_track:
        ax.text(0, 1, "North", horizontalalignment="center", verticalalignment="bottom")
        ax.text(0, -1, "South", horizontalalignment="center", verticalalignment="top")
        ax.text(1, 0, "East", horizontalalignment="left")
        ax.text(-1, 0, "West", horizontalalignment="right")

        ax.axis("off")
        ax.set_xlim([-1.1, 1.1])
        ax.set_ylim([-1.1, 1.1])
        ax.set_aspect("equal", "datalim")

    return fig, ax



import typing as t
import numpy.typing as npt
import pandas as pd
import pyproj
import lets_plot as lp
import bokeh.plotting as bp
import bokeh.models as bokeh_models
from sorts.types import EcefStates
from sorts.utils import assert_class_attributes_equal_to
from sorts.frames import ITRS_to_geodetic

EcefStatesPositionsPlotColumnKey = t.Literal["lat", "lon", "wmx", "wmy"]


class _K_EcefStatesPositionsPlotColumnKey:
    """Internal helper class for accessing string keys consistently"""

    lat: t.Final = "lat"
    lon: t.Final = "lon"
    wmx: t.Final = "wmx"
    wmy: t.Final = "wmy"


assert_class_attributes_equal_to(
    _K_EcefStatesPositionsPlotColumnKey, t.get_args(EcefStatesPositionsPlotColumnKey)
)


def _ecef_states_positions_plot_from_cds(source: bokeh_models.ColumnarDataSource):
    """An internal ver of `ecef_states_positions_plot` that takes a bokeh `ColumnDataSource`."""

    _K = _K_EcefStatesPositionsPlotColumnKey
    plot = bp.figure(
        x_axis_type="mercator",
        y_axis_type="mercator",
        match_aspect=True,
    )
    plot.add_tile("CartoDB Positron", retina=True)

    scatter = plot.scatter(x=_K.wmx, y=_K.wmy, source=source)

    # TODO: Implement a fix for the antimeridian wrapping issue
    # NOTE: We are not plotting the connecting line segments for now
    #   because bokeh does not handle correct wrapping of lines crossing antimeridian
    #   https://discourse.bokeh.org/t/bokeh-tile-antimeridian-problem/6978/3
    # plot.line(wmx, wmy)

    # add custom tooltips, only for the scatter plot/glyphs
    plot.add_tools(
        bokeh_models.HoverTool(
            renderers=[scatter],
            tooltips=[
                ("index", "$index"),
                ("data (lat, lon)", f"(@{_K.lat}, @{_K.lon})"),
            ],
        )
    )

    return plot


def _ecef_states_positions_plot_cds_cols(ecefs: EcefStates):
    geodetic_coords = ITRS_to_geodetic(ecefs[0], ecefs[1], ecefs[2])
    transformer = pyproj.Transformer.from_crs(
        "EPSG:4326", "EPSG:3857"
    )  # World Geodetic System to Web Mercator
    lat = geodetic_coords[0]
    lon = geodetic_coords[1]
    wmx, wmy = transformer.transform(lat, lon)

    cols: dict[EcefStatesPositionsPlotColumnKey, npt.NDArray] = {
        "lat": lat,
        "lon": lon,
        "wmx": wmx,
        "wmy": wmy,
    }

    return cols


# TODO: add down sampling? radar control slice are in milliseconds, while the simulation are in days or longer
def ecef_states_positions_plot(ecefs: EcefStates):
    cols = _ecef_states_positions_plot_cds_cols(ecefs)
    plot = _ecef_states_positions_plot_from_cds(
        source=bokeh_models.ColumnDataSource(t.cast(dict, cols))
    )

    return plot


# TODO: remove?
# TODO: add down sampling? radar control slice are in milliseconds, while the simulation are in days or longer
def ecef_states_positions_plot_letsplot(ecefs: EcefStates):
    """Returns a `Dash` app, use `.run()` method to run it."""
    # plotly alternative: https://plotly.com/python/lines-on-maps/

    geodetic_coords = ITRS_to_geodetic(ecefs[0], ecefs[1], ecefs[2])
    lat = geodetic_coords[0]
    lon = geodetic_coords[1]
    latlon_df = pd.DataFrame({"lat": lat, "lon": lon})
    latlon_df = latlon_df.reset_index()
    latlon_df["lat_head"] = latlon_df["lat"].shift(-1)
    latlon_df["lon_head"] = latlon_df["lon"].shift(-1)

    # alternatively, `geom_map` can be used instead of `geom_livemap`
    #   `world_countries = lp_geo_data.geocode_countries().get_boundaries(resolution=1)`
    #   `lp.geom_map(map=world_countries, projection="epsg3857", fill='gray', color='very_light_grey', size=0.1) # mercator projection`
    plot = (
        lp.ggplot(latlon_df, lp.aes(x="lon", y="lat"))
        # map background
        + lp.geom_livemap(projection="epsg3857", zoom=1)  # mercator projection
        # plot data points as point on map
        + lp.geom_point(size=0.5, tooltips=lp.layer_tooltips().line("#: @index, lat:^y, lon:^x"))
        # also connect the points using line segments for easier reading
        + lp.geom_segment(lp.aes(xend="lon_head", yend="lat_head"), size=0.2)
        # mark the orbit direction on the 1st data point
        + lp.geom_segment(
            lp.aes(xend="lon_head", yend="lat_head"),
            data={"lat": lat[0:1], "lon": lon[0:1], "lat_head": lat[1:2], "lon_head": lon[1:2]},
            size=1,
            arrow=lp.arrow(),
        )
        # other settings
        + lp.ggtitle("World Map (In Mercator projection)")
        + lp.ggsize(800, 600)
    )

    return plot



#!/usr/bin/env python

"""

"""

import numpy as np

from .scan import Scan


class Plane(Scan):
    """A uniform sampling of a horizontal plane."""

    def __init__(
        self,
        min_elevation=30.0,
        altitude=200e3,
        x_size=50e3,
        y_size=50e3,
        x_num=20,
        y_num=20,
        dwell=0.1,
        x_offset=0.0,
        y_offset=0.0,
    ):
        super().__init__(coordinates="enu")
        self._dwell = dwell
        self.min_elevation = min_elevation
        self.altitude = altitude

        self.x_size = x_size
        self.y_size = y_size
        self.x_num = x_num
        self.y_num = y_num

        k = np.empty((3, x_num * y_num), dtype=np.float64)

        xv, yv = np.meshgrid(
            np.linspace(-x_size * 0.5, x_size * 0.5, num=self.x_num, endpoint=True) + x_offset,
            np.linspace(-y_size * 0.5, y_size * 0.5, num=self.y_num, endpoint=True) + y_offset,
            sparse=False,
            indexing="ij",
        )

        k[0, :] = xv.flatten()
        k[1, :] = yv.flatten()
        k[2, :] = altitude

        k = k / np.linalg.norm(k, axis=0)

        min_z = np.sin(np.radians(min_elevation))
        k = k[:, k[2, :] >= min_z]
        self.num = k.shape[1]
        self.pointings = k

    def dwell(self, t=None):
        if t is None:
            return self._dwell
        else:
            if isinstance(t, float) or isinstance(t, int):
                return self._dwell
            else:
                return np.ones(t.shape, dtype=t.dtype) * self._dwell

    def min_dwell(self):
        return self._dwell

    def cycle(self):
        return self.num * self._dwell

    def pointing(self, t):
        ind = (np.mod(t / self.cycle(), 1) * self.num).astype(np.int64)
        return self.pointings[:, ind]



from __future__ import annotations
import logging, math, typing as t
from dataclasses import dataclass
import numpy as np
import numpy.typing as npt
import xarray as xr
from sorts import radar, scheduling
from sorts.const import min_datetime64_us
from sorts.radar import Station
from sorts.frames import enu_to_ecef, ecef_to_enu, sph_to_cart
from sorts.types import (
    Float_as_deg,
    Datetime64_us,
    Float64_as_m,
    EcefCoordinates,
    EnuCoordinates,
    Datetime_Like,
)
from sorts.utils import to_datetime64_us
from . import pointing_funcs
from .controller_base import ControllerBase

logger = logging.getLogger(__name__)


@dataclass(kw_only=True)
class ControllerState:
    start_time: Datetime64_us
    end_time: Datetime64_us
    tx_schedule_size: int
    tx_pointings_of_a_cycle: EnuCoordinates
    """NOTE: It may contain out of range pointings"""

    @classmethod
    def empty(cls) -> t.Self:
        return cls(
            start_time=min_datetime64_us,
            end_time=min_datetime64_us,
            tx_schedule_size=0,
            tx_pointings_of_a_cycle=np.empty((3, 0), dtype=np.float64),
        )


class FenceScanController(ControllerBase):
    """
    Generate schedule for a fence scaning pattern

    - The preferred way to create instances of this class is via its class methods (e.g. `TrackerController.from_space_object`).
    - This class serve as a frontend to the `State` type in this module
    """

    ControllerState = ControllerState
    """shortcut to module attribute"""

    def __init__(
        self,
        tx_station: Station,
        rx_stations: t.Sequence[Station],
        azimuth: Float_as_deg,
        min_elevation: Float_as_deg,
        pointings_per_cycle: int,
        scan_range: npt.NDArray[Float64_as_m],
        exp_detail: scheduling.ExperimentDetail,
        station_id_pairs: list[tuple[radar.StationId, radar.StationId]],
        state: ControllerState,
    ):
        """
        NOTE: This is intended as an internal constructor, please use the constructor methods to create instances.
        """

        self.tx_station = tx_station
        self.rx_stations = rx_stations
        self.azimuth = azimuth
        self.min_elevation = min_elevation
        self.pointings_per_cycle = pointings_per_cycle
        # TODO: probably make a new scan controller that can do optimal scan range selection based
        # on angle between the tx and rx beam (as the angle goes to 0 the range step goes to
        # infinity because the important parameter is the range overlap of the rx beam on the tx
        # beam)
        self.scan_range = scan_range
        self.exp_detail = exp_detail
        self.station_id_pairs = station_id_pairs

        self.state: ControllerState = state

    @classmethod
    def from_scan_spec(
        cls,
        tx_station: Station,
        rx_stations: t.Sequence[Station],
        azimuth: Float_as_deg,
        min_elevation: Float_as_deg,
        pointings_per_cycle: int,
        scan_range: npt.NDArray[Float64_as_m],
        exp_detail: scheduling.ExperimentDetail,
    ) -> t.Self:
        """A constructor method"""

        stn_pairs = [(tx_station.uid, rx_station.uid) for rx_station in rx_stations]

        ctrl = cls(
            tx_station=tx_station,
            rx_stations=rx_stations,
            azimuth=azimuth,
            min_elevation=min_elevation,
            pointings_per_cycle=pointings_per_cycle,
            scan_range=scan_range,
            exp_detail=exp_detail,
            station_id_pairs=stn_pairs,
            state=ControllerState.empty(),
        )

        return ctrl

    def get_experiment_detail(self) -> scheduling.ExperimentDetail:
        return self.exp_detail

    def get_experiment_id_station_id_pairs_map(self) -> scheduling.ExperimentIdStationIdPairsMap:
        return {self.exp_detail.id: self.station_id_pairs}

    def get_station_map(self) -> dict[radar.StationId, radar.Station]:
        stn_map: dict[radar.StationId, radar.Station] = {}

        stn_map[self.tx_station.uid] = self.tx_station
        stn_map.update(list([(stn.uid, stn) for stn in self.rx_stations]))

        return stn_map

    def _compute_controller_state(
        self, start_time: Datetime_Like, end_time: Datetime_Like
    ) -> ControllerState:
        """Do the computation and return the updated `state` property."""

        exp_detail = self.exp_detail

        start_time_np = to_datetime64_us(start_time)
        end_time_np = to_datetime64_us(end_time)
        tx_schedule_size = math.floor((end_time_np - start_time_np) / exp_detail.slice_duration)

        tx_pointings_of_a_cycle = sph_to_cart(
            pointing_funcs.fence_pattern(
                azimuth=self.azimuth,
                min_elevation=self.min_elevation,
                pointings_per_cycle=self.pointings_per_cycle,
            ),
            degrees=True,
        )

        state = ControllerState(
            start_time=start_time_np,
            end_time=end_time_np,
            tx_schedule_size=tx_schedule_size,
            tx_pointings_of_a_cycle=tx_pointings_of_a_cycle,
        )

        return state

    def _generate(self) -> scheduling.Schedule:
        """Generate the schedules."""

        # TODO: write schedule to disk generally and then chunk load it as needed in the actual
        # simulation, the general simulation pattern will be "1. propagate objects and generate
        # states and make schedule, 2. run simulation, 3. analyze results"

        # The logic of this function:
        # 1. repeat the cycle of tx pointings from state to form the tx schedule
        # 2. from the single cycle of tx pointings, we convert it into ECEF location coord and extend them by the `scan_range`
        # 3. using the resultant location coords from previous step,
        #    we convert them to rx station pointings of a cycle in ECEF coord,
        #    and then further back to pointings in ENU coord,
        #    and finally repeat them to form a rx schedule, for each rx station

        pointings_per_cycle = self.state.tx_pointings_of_a_cycle.shape[1]

        # NOTE: for `np.arange` 'stop param,
        #   - we subtract 'slice_duration' so that only full slice are included
        #   - and add `+1` so that slice with time range `('end_time - 'slice_duration', 'end_time')` is included
        tx_slice_start_time: npt.NDArray[Datetime64_us] = np.arange(
            self.state.start_time,
            self.state.end_time - self.exp_detail.slice_duration + 1,
            self.exp_detail.slice_duration,
        )

        # TODO: `tx_schedule_size` is a bit of a mismisnomer, as out-of-range entries might later be removed
        # repeat a cycle of pointings until it is at least the size of `tx_schedule_size`,
        # then trim to exactly `tx_schedule_size` long
        tx_pointing: EnuCoordinates = np.tile(
            self.state.tx_pointings_of_a_cycle,
            (self.state.tx_schedule_size + pointings_per_cycle - 1) // pointings_per_cycle,
        )[:, : self.state.tx_schedule_size]

        # mask tx values by min_elevation requirement
        tx_mask = pointing_funcs.create_mask_by_min_elevation(
            tx_pointing, self.tx_station.min_elevation
        )
        tx_slice_start_time_masked = tx_slice_start_time[tx_mask]
        tx_pointing_masked = tx_pointing[:, tx_mask]

        tx_sch = scheduling.from_ndarrays(
            start_time=tx_slice_start_time_masked,
            end_time=tx_slice_start_time_masked + self.exp_detail.slice_duration,
            exp_num=np.full(len(tx_slice_start_time_masked), self.exp_detail.id, dtype=np.int16),
            stn_num=np.full(len(tx_slice_start_time_masked), self.tx_station.uid, dtype=np.int16),
            simult_num=np.full(len(tx_slice_start_time_masked), 0, dtype=np.int16),
            pointing=tx_pointing_masked,
        )

        # TODO: `rx_schedule_size` is a bit of a mismisnomer, as out-of-range entries might later be removed
        rx_slice_start_time = tx_slice_start_time.repeat(len(self.scan_range))
        rx_schedule_size = self.state.tx_schedule_size * len(self.scan_range)
        rx_schs: list[scheduling.Schedule] = []
        tx_pointings_of_a_cycle_without_translation_ecef: EcefCoordinates = enu_to_ecef(
            lat=self.tx_station.ecef_lat,
            lon=self.tx_station.ecef_lon,
            alt=self.tx_station.ecef_alt,
            enu=self.state.tx_pointings_of_a_cycle,
            degrees=True,
        )
        rx_pointing_of_a_cycle_ecef: EcefCoordinates = (
            tx_pointings_of_a_cycle_without_translation_ecef[:, :, np.newaxis]
            * self.scan_range[np.newaxis, np.newaxis, :]
            + self.tx_station.ecef[:, np.newaxis, np.newaxis]
        ).reshape((3, -1))

        for rx_station in self.rx_stations:
            rx_pointings_of_a_cycle_without_translation_ecef: EcefCoordinates = (
                rx_pointing_of_a_cycle_ecef - rx_station.ecef[:, np.newaxis]
            )
            rx_pointings_of_a_cycle_enu: EnuCoordinates = ecef_to_enu(
                lat=rx_station.ecef_lat,
                lon=rx_station.ecef_lon,
                alt=rx_station.ecef_alt,
                ecef=rx_pointings_of_a_cycle_without_translation_ecef,
                degrees=True,
            )

            # repeat a cycle of pointings until it is at least the size of `rx_schedule_size`
            # then trim to exactly `rx_schedule_size` long
            rx_pointings_enu: EnuCoordinates = np.tile(
                rx_pointings_of_a_cycle_enu,
                (rx_schedule_size + pointings_per_cycle - 1) // pointings_per_cycle,
            )[:, :rx_schedule_size]
            rx_pointings_simult_num = np.arange(rx_schedule_size) % len(self.scan_range)

            # mask rx values by min_elevation requirement, and has a corresponding tx value
            rx_mask_by_min_elevation = pointing_funcs.create_mask_by_min_elevation(
                rx_pointings_enu, rx_station.min_elevation
            )
            rx_mask_by_tx_mask = np.isin(rx_slice_start_time, tx_slice_start_time_masked)
            rx_mask = np.logical_and(rx_mask_by_min_elevation, rx_mask_by_tx_mask)
            rx_slice_start_time_masked = rx_slice_start_time[rx_mask]
            rx_pointing_masked = rx_pointings_enu[:, rx_mask]
            rx_pointings_simult_num_masked = rx_pointings_simult_num[rx_mask]

            rx_sch = scheduling.from_ndarrays(
                start_time=rx_slice_start_time_masked,
                end_time=rx_slice_start_time_masked + self.exp_detail.slice_duration,
                exp_num=np.full(
                    len(rx_slice_start_time_masked), self.exp_detail.id, dtype=np.int16
                ),
                stn_num=np.full(len(rx_slice_start_time_masked), rx_station.uid, dtype=np.int16),
                simult_num=rx_pointings_simult_num_masked,
                pointing=rx_pointing_masked,
            )

            rx_schs.append(rx_sch)

        resultant_sch = xr.concat([tx_sch, *rx_schs], dim=scheduling._K.multi_index)
        resultant_sch = resultant_sch.sortby(scheduling._K.start_time)
        # TODO: re-eval if it is too brutal
        # there will be duplicates if the tx station is also a rx station, we drop the duplicates here
        resultant_sch = resultant_sch.drop_duplicates(scheduling._K.multi_index)
        output = resultant_sch

        return output

    def generate(self, start_time: Datetime_Like, end_time: Datetime_Like) -> scheduling.Schedule:
        """Generate the schedules."""

        self.state = self._compute_controller_state(start_time, end_time)
        sch = self._generate()

        return sch



#!/usr/bin/env python

"""Ionospheric radio propagation effects studied using ray-tracing.

2016-2018 Juha Vierinen
2018-2020 Juha Vierinen, Daniel Kastinen

"""

try:
    from pyglow.pyglow import Point
    from pyglow import coord as gcoord
except ImportError:
    Point = None
    gcoord = None

from astropy.time import Time

import numpy as np
import scipy.integrate as integrate
import scipy.interpolate as interpolate
import scipy.constants as constants

from .. import frames
from .errors import Errors


def calculate_delay(
    time,
    lat,
    lon,
    frequency,
    elevation,
):
    """TODO: Docstring"""

    if Point is None or gcoord is None:
        raise ImportError("pyglow must be installed to calculate delay")

    if not isinstance(time, Time):
        dn = time
    else:
        dn = time.tt.datetime

    num = 500
    alts = np.linspace(0, 4000, num=num)
    distance = np.linspace(0, 4000, num=num)
    ne = np.zeros(num)
    xyz_prev = 0.0
    for ai, a in enumerate(alts):
        llh = coord.az_el_r2geodetic(lat, lon, 0, 180.0, elevation, a * 1e3)

        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()
        if pt.ne > 0.0:
            ne[ai] = pt.ne * 1e6
        else:
            ne[ai] = 0.0
        xyz = gcoord.lla2ecef(np.array([lat, lon, a]))[0]
        if ai == 0:
            distance[ai] = 0.0
        else:
            distance[ai] = np.sqrt(np.dot(xyz - xyz_prev, xyz - xyz_prev)) + distance[ai - 1]
        xyz_prev = xyz

    f_p = 8.98 * np.sqrt(ne)
    v_g = constants.c * np.sqrt(1 - (f_p / frequency) ** 2.0)

    dt2 = (
        integrate.simps(1.0 - 1.0 / (np.sqrt(1 - (f_p / frequency) ** 2.0)), distance * 1e3)
        / constants.c
    )

    return dt2, ne, distance


def ray_trace(
    time,
    lat,
    lon,
    frequency,
    elevation,
    azimuth,
):
    """TODO: Docstring"""

    if Point is None or gcoord is None:
        raise ImportError("pyglow must be installed to ray trace")

    if not isinstance(time, Time):
        dn = time
    else:
        dn = time.tt.datetime

    num = 1000
    alts = np.linspace(0, 4000, num=num)
    distance = np.linspace(0, 4000, num=num)
    ne = np.zeros(num)
    ne2 = np.zeros(num)
    dnex = np.zeros(num)
    dtheta = np.zeros(num)
    dalt = np.zeros(num)
    dney = np.zeros(num)
    dnez = np.zeros(num)
    xyz_prev = 0.0
    px = np.zeros(num)
    dk = np.zeros(num)
    py = np.zeros(num)
    pz = np.zeros(num)
    p0x = np.zeros(num)
    p0y = np.zeros(num)
    p0z = np.zeros(num)

    # initial direction and position
    k = frames.azel_to_ecef(lat, lon, 10e3, azimuth, elevation)
    k0 = k
    p = frames.geodetic_to_ITRS(lat, lon, 10e3)
    pe = frames.geodetic_to_ITRS(lat, lon, 10e3)
    p0 = frames.geodetic_to_ITRS(lat, lon, 10e3)
    dh = 4e3
    vg = 1.0

    p_orig = p
    ray_time = 0.0

    for ai, a in enumerate(alts):
        p = p + k * dh * vg
        p0 = p0 + k0 * dh
        ray_time += dh / constants.c

        dpx = p + np.array([1.0, 0.0, 0.0]) * dh
        dpy = p + np.array([0.0, 1.0, 0.0]) * dh
        dpz = p + np.array([0.0, 0.0, 1.0]) * dh

        llh = frames.ITRS_to_geodetic(p[0], p[1], p[2])
        llh_1 = frames.ITRS_to_geodetic(p0[0], p0[1], p0[2])
        dalt[ai] = llh_1[2] - llh[2]

        if llh[2] / 1e3 > 1900:
            break
        alts[ai] = llh[2] / 1e3
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()

        if pt.ne > 0.0:
            ne[ai] = pt.ne * 1e6
            f_p = 8.98 * np.sqrt(ne[ai])
            v_g = np.sqrt(1.0 - (f_p / frequency) ** 2.0)
        else:
            ne[ai] = 0.0

        llh = frames.ITRS_to_geodetic(dpx[0], dpx[1], dpx[2])
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()
        if pt.ne > 0.0:
            dnex[ai] = (ne[ai] - pt.ne * 1e6) / dh
        else:
            dnex[ai] = 0.0

        llh = frames.ITRS_to_geodetic(dpy[0], dpy[1], dpy[2])
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()
        if pt.ne > 0.0:
            dney[ai] = (ne[ai] - pt.ne * 1e6) / dh
        else:
            dney[ai] = 0.0

        llh = frames.ITRS_to_geodetic(dpz[0], dpz[1], dpz[2])
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()
        if pt.ne > 0.0:
            dnez[ai] = (ne[ai] - pt.ne * 1e6) / dh
        else:
            dnez[ai] = 0.0
        grad = np.array([dnex[ai], dney[ai], dnez[ai]])
        px[ai] = p[0]
        py[ai] = p[1]
        pz[ai] = p[2]
        p0x[ai] = p0[0]
        p0y[ai] = p0[1]
        p0z[ai] = p0[2]

        dk[ai] = np.arccos(np.dot(k0, k) / (np.sqrt(np.dot(k0, k0)) * np.sqrt(np.dot(k, k))))
        # no bending if gradient too small
        if np.dot(grad, grad) > 100.0:
            grad1 = grad / np.sqrt(np.dot(grad, grad))

            p2 = p + k * dh
            llh = frames.ITRS_to_geodetic(p2[0], p2[1], p2[2])
            pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
            pt.run_iri()
            if pt.ne > 0.0:
                ne2 = pt.ne * 1e6
            else:
                ne2 = 0.0
            f0 = 8.98 * np.sqrt(ne[ai])
            n0 = np.sqrt(1.0 - (f0 / frequency) ** 2.0)
            f1 = 8.98 * np.sqrt(ne2)
            n1 = np.sqrt(1.0 - (f1 / frequency) ** 2.0)

            theta0 = np.arccos(
                np.dot(grad, k) / (np.sqrt(np.dot(grad, grad)) * np.sqrt(np.dot(k, k)))
            )
            # angle cannot be over 90
            if theta0 > np.pi / 2.0:
                theta0 = np.pi - theta0
            sin_theta_1 = (n0 / n1) * np.sin(theta0)
            dtheta[ai] = 180.0 * np.arcsin(sin_theta_1) / np.pi - 180.0 * theta0 / np.pi
            # print("n0/n1 %1.10f theta0 %1.2f theta1-theta0 %1.10f"%(n0/n1,180.0*theta0/np.pi,dtheta[ai]))
            cos_theta_1 = np.sqrt(1.0 - sin_theta_1**2.0)
            k_ref = (n0 / n1) * k + ((n0 / n1) * np.cos(theta0) - cos_theta_1) * grad1
            # normalize
            k_ref / np.sqrt(np.dot(k_ref, k_ref))
            k = k_ref

            angle = np.arccos(np.dot(grad, k) / np.sqrt(np.dot(grad, grad)) * np.sqrt(np.dot(k, k)))

    los_time = np.sqrt(np.dot(p_orig - p, p_orig - p)) / constants.c
    excess_ionospheric_delay = ray_time - los_time

    # print("Excess propagation time %1.20f mus"%((1e6*(ray_time-los_time))))

    theta = np.arccos(np.dot(k0, k) / (np.sqrt(np.dot(k0, k0)) * np.sqrt(np.dot(k, k))))

    theta_p = np.arccos(np.dot(p0, p) / (np.sqrt(np.dot(p0, p0)) * np.sqrt(np.dot(p, p))))

    llh0 = frames.ITRS_to_geodetic(px[ai - 2], py[ai - 2], pz[ai - 2])
    llh1 = frames.ITRS_to_geodetic(p0x[ai - 2], p0y[ai - 2], p0z[ai - 2])

    # print("d_coord")
    # print(llh0-llh1)

    ret = dict(
        px=px,
        py=py,
        pz=pz,
        p0x=p0x,
        p0y=p0y,
        p0z=p0z,
        ray_bending=dtheta,
        electron_density=ne,
        altitudes=alts,
        altitude_errors=dalt,
        excess_ionospheric_delay=excess_ionospheric_delay,
        total_angle_error=180.0 * theta_p / np.pi,
        p_end=p,
        p0_end=p0,
    )

    return ret


def ray_trace_error(
    time,
    lat,
    lon,
    frequency,
    elevation,
    azimuth,
    ionosphere=False,
    error_std=0.05,
):
    """TODO: Docstring"""

    if Point is None or gcoord is None:
        raise ImportError("pyglow must be installed to ray trace")

    if not isinstance(time, Time):
        dn = time
    else:
        dn = time.tt.datetime

    num = 2000
    alts = np.repeat(1e99, num)
    distance = np.linspace(0, 4000, num=num)
    ne = np.zeros(num)
    ne2 = np.zeros(num)
    dtheta = np.zeros(num)
    dalt = np.zeros(num)
    dnex = np.zeros(num)
    dney = np.zeros(num)
    dnez = np.zeros(num)
    xyz_prev = 0.0
    dk = np.zeros(num)
    px = np.zeros(num)
    py = np.zeros(num)
    pz = np.zeros(num)
    t_vec = np.zeros(num)
    t_i_vec = np.zeros(num)
    k_vecs = []
    # initial direction and position
    k = frames.azel_to_ecef(lat, lon, 10e3, az, elevation)
    k0 = k
    p = frames.geodetic_to_ITRS(lat, lon, 10e3)
    dh = 4e3
    dt = 20e-6
    # correlated errors std=1, 100 km correlation length
    scale_length = 40.0
    ne_errors_x = np.convolve(
        np.repeat(1.0 / np.sqrt(scale_length), scale_length), np.random.randn(10000)
    )
    ne_errors_y = np.convolve(
        np.repeat(1.0 / np.sqrt(scale_length), scale_length), np.random.randn(10000)
    )
    ne_errors_z = np.convolve(
        np.repeat(1.0 / np.sqrt(scale_length), scale_length), np.random.randn(10000)
    )

    p_orig = p
    ray_time = 0.0
    v_c = constants.c
    for ai, a in enumerate(alts):
        # go forward in time
        dhp = v_c * dt
        p = p + k * dhp
        ray_time += dt
        print(ray_time * 1e6)
        t_vec[ai + 1] = dt
        k_vecs.append(k)

        dpx = p + np.array([1.0, 0.0, 0.0]) * dh
        dpy = p + np.array([0.0, 1.0, 0.0]) * dh
        dpz = p + np.array([0.0, 0.0, 1.0]) * dh

        llh = frames.ITRS_to_geodetic(p[0], p[1], p[2])

        if llh[2] / 1e3 > 2100:
            break
        alts[ai] = llh[2] / 1e3
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()

        if pt.ne > 0.0:
            ne[ai] = pt.ne * (1.0 + error_std * ne_errors_x[ai]) * 1e6
            if ionosphere:
                f0 = 8.98 * np.sqrt(ne[ai])
                f_p = 8.98 * np.sqrt(ne[ai])
                # update group velocity
                v_c = constants.c * np.sqrt(1.0 - (f0 / frequency) ** 2.0)
        else:
            ne[ai] = 0.0

        llh = frames.ITRS_to_geodetic(dpx[0], dpx[1], dpx[2])
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()

        if pt.ne > 0.0:
            dnex[ai] = (ne[ai] - pt.ne * (1.0 + error_std * ne_errors_x[ai]) * 1e6) / dh
        else:
            dnex[ai] = 0.0

        llh = frames.ITRS_to_geodetic(dpy[0], dpy[1], dpy[2])
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()

        if pt.ne > 0.0:
            dney[ai] = (ne[ai] - pt.ne * (1.0 + error_std * ne_errors_x[ai]) * 1e6) / dh
        else:
            dney[ai] = 0.0

        llh = frames.ITRS_to_geodetic(dpz[0], dpz[1], dpz[2])
        pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
        pt.run_iri()

        if pt.ne > 0.0:
            dnez[ai] = (ne[ai] - pt.ne * (1.0 + error_std * ne_errors_x[ai]) * 1e6) / dh
        else:
            dnez[ai] = 0.0

        grad = np.array([dnex[ai], dney[ai], dnez[ai]])

        px[ai] = p[0]
        py[ai] = p[1]
        pz[ai] = p[2]

        dk[ai] = np.arccos(np.dot(k0, k) / (np.sqrt(np.dot(k0, k0)) * np.sqrt(np.dot(k, k))))
        # no bending if gradient too small
        if np.dot(grad, grad) > 100.0 and ionosphere:
            grad1 = grad / np.sqrt(np.dot(grad, grad))

            p2 = p + k * dh
            llh = frames.ITRS_to_geodetic(p2[0], p2[1], p2[2])
            pt = Point(dn, llh[0], llh[1], llh[2] / 1e3)
            pt.run_iri()
            if pt.ne > 0.0:
                ne2 = pt.ne * (1.0 + error_std * ne_errors_x[ai]) * 1e6
            else:
                ne2 = 0.0
            f0 = 8.98 * np.sqrt(ne[ai])
            n0 = np.sqrt(1.0 - (f0 / frequency) ** 2.0)
            f1 = 8.98 * np.sqrt(ne2)
            n1 = np.sqrt(1.0 - (f1 / frequency) ** 2.0)

            theta0 = np.arccos(
                np.dot(grad, k) / (np.sqrt(np.dot(grad, grad)) * np.sqrt(np.dot(k, k)))
            )
            # angle cannot be over 90
            if theta0 > np.pi / 2.0:
                theta0 = np.pi - theta0
            sin_theta_1 = (n0 / n1) * np.sin(theta0)
            dtheta[ai] = 180.0 * np.arcsin(sin_theta_1) / np.pi - 180.0 * theta0 / np.pi
            #            print("n0/n1 %1.10f theta0 %1.2f theta1-theta0 %1.10f"%(n0/n1,180.0*theta0/np.pi,dtheta[ai]))
            cos_theta_1 = np.sqrt(1.0 - sin_theta_1**2.0)
            k_ref = (n0 / n1) * k + ((n0 / n1) * np.cos(theta0) - cos_theta_1) * grad1
            # normalize
            k_ref / np.sqrt(np.dot(k_ref, k_ref))
            k = k_ref

            angle = np.arccos(np.dot(grad, k) / np.sqrt(np.dot(grad, grad)) * np.sqrt(np.dot(k, k)))

    return t_vec, px, py, pz, alts, ne, k_vecs


def ionospheric_error(time, elevation=90.0, n_samp=20, frequency=233e6, error_std=0.05):
    """TODO: Docstring

    # estimate using sampling what the ray-tracing error is
    # return error in microseconds, round-trip time units.
    """
    prop_errors = np.zeros([n_samp, 100])
    prop_error_mean = np.zeros(100)
    prop_error_std = np.zeros(100)
    prop_alts = np.linspace(0, 2000, num=100)

    for i in range(n_samp):
        t_vec_i, px_i, py_i, pz_i, alts_i, ne_i, k_i = ray_trace_error(
            time=time,
            frequency=frequency,
            elevation=elevation,
            ionosphere=True,
            error_std=0.0,
        )
        t_vec_e_i, px_e_i, py_e_i, pz_e_i, alts_e_i, ne_e_i, k_e_i = ray_trace_error(
            time=time,
            frequency=frequency,
            elevation=elevation,
            ionosphere=True,
            error_std=error_std,
        )

        maxi = np.where(alts_i > 2050)[0][0]
        alts_i[0] = 0.0

        offsets = np.zeros(maxi)
        for ri in range(maxi):
            # determine what is the additional distance needed to reach target
            p = np.array([px_i[ri], py_i[ri], pz_i[ri]])
            pe = np.array([px_e_i[ri], py_e_i[ri], pz_e_i[ri]])

            offsets[ri] = (
                1e6
                * 2.0
                * np.abs(np.dot(p - pe, k_i[ri] / np.sqrt(np.dot(k_i[ri], k_i[ri]))))
                / constants.c
            )
        pos_err_fun = interpolate.interp1d(alts_i[0:maxi], offsets)
        prop_errors[i, :] = pos_err_fun(prop_alts)

    for ri in range(len(prop_alts)):
        prop_error_mean[ri] = np.mean(prop_errors[:, ri])
        prop_error_std[ri] = np.std(prop_errors[:, ri])

    return (
        prop_error_mean,
        prop_error_std / np.sqrt(float(n_samp)),
        prop_alts,
        ne_i[0:maxi],
        ne_e_i[0:maxi],
        alts_i[0:maxi],
    )


class IonosphericRayTrace(Errors):
    """ """

    VARIABLES = [
        "range",
        "k",
        "t",
    ]

    def __init__(self, station, seed=None, electron_density_std=0.0):
        super().__init__(seed=seed)

        self.electron_density_std = electron_density_std
        self.station = station

    def k(self, data, time, azimuth, elevation):
        raise NotImplementedError("")

    def t(self, data, time, azimuth, elevation):
        raise NotImplementedError("")

    def range(self, data, time, azimuth, elevation):
        raise NotImplementedError("")
        ray_trace_error(
            time,
            self.station.lat,
            self.station.lon,
            self.station.beam.frequency,
            elevation,
            azimuth,
            ionosphere=True,
            error_std=self.electron_density_std,
        )



#!/usr/bin/env python

"""Functions that concern dates and times.

"""

# Python standard import
import time
import datetime
import os

# Third party import
import numpy as np

# Local import

os.environ["TZ"] = "GMT"
time.tzset()


def date_to_unix(year, month, day, hour, minute, second):
    """Convert date to unix time in seconds


    :param int year: Year as integer. Years preceding 1 A.D. should be 0 or negative. The year before 1 A.D. is 0, 10 B.C. is year -9.
    :param int month: Month as integer, Jan = 1, Feb. = 2, etc.
    :param int day:  Day
    :param int hour:  Hour in 24h format
    :param int minute:  Minute
    :param float second:  Second, may contain fractional part.
    :return: Unix time in seconds
    :rtype: float
    """
    t = datetime.datetime(year, month, day, hour, minute, second)
    return time.mktime(t.timetuple())


def unix_to_date(unix):
    """Convert unix time in seconds to UTC date datetime object

    :param float unix: Unix time in seconds.
    :return: Datetime object in UTC
    :rtype: datetime.datetime
    """
    return datetime.datetime.utcfromtimestamp(unix)


def unix_to_datestr(unix):
    """Convert unix time in seconds to Gregorian calendar UTC date-time formatted string

    :param float unix: Unix time in seconds.
    :return: Gregorian calendar UTC date-time formatted string
    :rtype: str
    """
    return unix_to_date(unix).strftime("%Y-%m-%dT%H:%M:%S")


def unix_to_datestrf(x):
    """Convert unix time in seconds to Gregorian calendar UTC date-time formatted string

    Different implementation?

    :param float unix: Unix time in seconds.
    :return: Gregorian calendar UTC date-time formatted string
    :rtype: str
    """
    return "%s" % (unix_to_date(x).strftime("%Y-%m-%dT%H:%M:%S.%f"))


sec = np.timedelta64(1000000000, "ns")
"""numpy.datetime64: Interval of 1 second
"""


def jd_to_mjd(jd):
    """Convert Julian Date (relative 12h Jan 1, 4713 BC) to Modified Julian Date (relative 0h Nov 17, 1858)"""
    return jd - 2400000.5


def mjd_to_jd(mjd):
    """Convert Modified Julian Date (relative 0h Nov 17, 1858) to Julian Date (relative 12h Jan 1, 4713 BC)"""
    return mjd + 2400000.5


def mjd_to_j2000(mjd_tt):
    """Convert from Modified Julian Date to days past J2000.

    :param float/numpy.ndarray mjd_tt: MJD in TT
    :return: Days past J2000
    :rtype: float/numpy.ndarray
    """
    return mjd_to_jd(mjd_tt) - 2451545.0


def npdt_to_date(dt):
    """Converts a numpy datetime64 value to a date tuple

    :param numpy.datetime64 dt: Date and time (UTC) in numpy datetime64 format

    :return: tuple (year, month, day, hours, minutes, seconds, microsecond)
             all except usec are integer
    """

    t0 = np.datetime64("1970-01-01", "s")
    ts = (dt - t0) / sec

    it = int(np.floor(ts))
    usec = 1e6 * (dt - (t0 + it * sec)) / sec

    tm = time.localtime(it)
    return tm.tm_year, tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, usec


def npdt_to_mjd(dt):
    """Converts a numpy datetime64 value (UTC) to a modified Julian date"""
    return (dt - np.datetime64("1858-11-17")) / np.timedelta64(1, "D")


def mjd_to_npdt(mjd):
    """Converts a modified Julian date to a numpy datetime64 value (UTC)"""
    day = np.timedelta64(24 * 3600 * 1000 * 1000, "us")
    return np.datetime64("1858-11-17") + day * mjd


def unix_to_npdt(unix):
    """Converts unix seconds to a numpy datetime64 value (UTC)"""
    return np.datetime64("1970-01-01") + np.timedelta64(1000 * 1000, "us") * unix


def npdt_to_unix(dt):
    """Converts a numpy datetime64 value (UTC) to unix seconds"""
    return (dt - np.datetime64("1970-01-01")) / np.timedelta64(1, "s")



import logging


# NOTE: here are some notes from danielk on logger setup/config
#   we should see if the way we configure logging by default is good or not and refactor that as
#   needed, e.g. if we look at
#   https://github.com/danielk333/runningman/blob/main/src/runningman/manager.py#L125
#   this is how i setup logging there, i think this code is similar
#   We should aim for the pattern of: if the user wants to log they can setup logging themself or
#   use a convenience function that does `setup_logging(bunch of useful parameters)` and it will configure
#   most things
def apply_suggested_config(logger: logging.Logger):
    """configure a logger to a suggested configuration"""

    handler = logging.StreamHandler()  # this goes to console
    formatter = logging.Formatter("[%(asctime)s]%(levelname)s:%(name)s:%(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    logger.setLevel(logging.INFO)



#!/usr/bin/env python

"""Defines grid sampled populations over some parameters."""
from typing import Any
import numpy as np
from astropy.time import Time
from .population import Population
from sorts.types import NDArray_N, Frames


def orbit_grid(
    semi_major_axis_samples: NDArray_N,
    eccentricity_samples: NDArray_N,
    inclination_samples: NDArray_N,
    argument_of_periapsis_samples: NDArray_N,
    longitude_of_ascending_node_samples: NDArray_N,
    mean_anomaly_samples: NDArray_N,
    diameter_samples: NDArray_N,
    frame: Frames = "GCRS",
    epoch_mjd: NDArray_N | float = 53005.0,
    additional_parameters: dict[str, Any] = None,
    degrees: bool = True,
):
    samples = [
        semi_major_axis_samples,
        eccentricity_samples,
        inclination_samples,
        argument_of_periapsis_samples,
        longitude_of_ascending_node_samples,
        mean_anomaly_samples,
        diameter_samples,
    ]
    grids = [x.flatten() for x in np.meshgrid(*samples)]
    size = grids[0].size
    parameters = {"d": grids[-1]}
    if additional_parameters is not None:
        parameters.update(additional_parameters)

    pop = Population(
        states=np.stack(grids[:-1]),
        epochs=Time(
            np.full((size,), epoch_mjd, dtype=np.float64),
            format="mjd",
            scale="utc",
        ),
        frame=frame,
        parameters=parameters,
        object_ids=np.arange(size),
        state_format="kepler",
        anomly_type="mean",
        dtypes={"id": np.int64},
        default_dtype=np.float64,
        epoch_format="mjd",
        epoch_scale="utc",
        degrees=degrees,
    )
    return pop



import logging
import numpy as np
import numpy.typing as npt
import spacecoords
from sorts.types import Float_as_deg, AzelrCoordinates_DegM, EnuCoordinates

logger = logging.getLogger(__name__)


def fence_pattern(
    azimuth: Float_as_deg,
    min_elevation: Float_as_deg,
    pointings_per_cycle: int,
) -> AzelrCoordinates_DegM:
    """
    Generate radar pointings that evenly sweep over a symmetrical elevation range, at the given `azimuth`

    Note that the sweeping always strokes in the same direction (not back and forth).
    """

    el = np.linspace(
        min_elevation, 180.0 - min_elevation, num=pointings_per_cycle, dtype=np.float64
    )
    az = np.full(pointings_per_cycle, azimuth, dtype=np.float64)

    # make 0 <= el < 90
    el_over_90deg_mask = el > 90.0
    el[el_over_90deg_mask] = 180.0 - el[el_over_90deg_mask]

    # wrap around az for those with el > 90 deg
    az[el_over_90deg_mask] = np.mod(az[el_over_90deg_mask] + 180.0, 360.0)

    azelr = np.stack(
        [
            az,
            el,
            np.full(pointings_per_cycle, 1.0, dtype=np.float64),
        ],
    )

    return azelr


def create_mask_by_min_elevation(
    pointings: EnuCoordinates, min_elevation: float
) -> npt.NDArray[np.bool]:
    loc_zenith = np.array([0, 0, 1], dtype=np.float64)

    _pointings_zenith_ang = spacecoords.linalg.vector_angle(loc_zenith, pointings, degrees=True)
    match _pointings_zenith_ang:
        case np.ndarray():
            pointings_zenith_ang = _pointings_zenith_ang
        case float():
            pointings_zenith_ang = np.array([_pointings_zenith_ang], dtype=np.float64)
        case _:
            raise RuntimeError(
                f"unexpected type of `_pointings_zenith_ang`: {type(_pointings_zenith_ang)}"
            )

    el_in_range_mask = pointings_zenith_ang <= 90.0 - min_elevation

    return el_in_range_mask



#!/usr/bin/env python

"""

"""

import numpy as np

from .fence import Fence


class RandomFence(Fence):
    """General fence scan."""

    def pointing(self, t):
        raise NotImplementedError()

        ind = (np.mod(t / self.cycle(), 1) * self.num).astype(np.int64)
        if isinstance(t, float) or isinstance(t, int):
            shape = (3,)
        else:
            shape = (3, len(t))

        azelr = np.empty(shape, dtype=np.float64)
        azelr[0, ...] = self._az[ind]
        azelr[1, ...] = self._el[ind]
        azelr[2, ...] = 1.0
        return azelr
